/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
//import java.util.LinkedHashSet;
import uk.ac.leeds.ccg.andyt.grids.core.AbstractGrid2DSquareCell.ChunkID;

/**
 * A class to be extended so as to handle OutOfMemoryErrors.
 */
public abstract class OutOfMemoryErrorHandler
        implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    /**
     * A HashSet of AbstractGrid2DSquareCell objects that
     * have data that can be swapped to release memory for
     * processing.
     */
    protected transient HashSet _Grid2DSquareCells;
    //protected transient LinkedHashSet _Grid2DSquareCells;
    
    protected transient Runtime _Runtime;
    
    /**
     * For ease of search and replace coding.
     */
    public static final boolean _HandleOutOfMemoryErrorFalse = false;
    
    /**
     * For ease of search and replace coding.
     */
    public static final boolean _HandleOutOfMemoryErrorTrue = true;
    
    /**
     * Reserve memory that can be set to null and garbage collected
     * so as to handle OutOfMemoryErrors.
     */
    protected transient int[] _MemoryReserve;
    //protected transient int[] _MemoryReserve;
    
    /**
     * Initialises _Runtime
     */
    protected final void init_Runtime() {
        this._Runtime = Runtime.getRuntime();
    }
    
    /**
     * Initialises _Grid2DSquareCells
     */
    protected void init_Grid2DSquareCells() {
        if ( this._Grid2DSquareCells == null ) {
            this._Grid2DSquareCells = new HashSet();
        }
    }
    
    /**
     * Initialises _Grid2DSquareCells
     */
    protected void init_Grid2DSquareCells(
            HashSet _Grid2DSquareCells ) {
        if ( this._Grid2DSquareCells == null ) {
            this._Grid2DSquareCells = _Grid2DSquareCells;
        } else {
            //System.err.println( this.getClass().getName() + ".init_Grid2DSquareCells(HashSet)" );
            if ( _Grid2DSquareCells == null ) { // Debug
                this._Grid2DSquareCells = new HashSet();
            } else {
                this._Grid2DSquareCells = _Grid2DSquareCells;
            }
        }
    }
    
    /**
     * @return this._Grid2DSquareCells.
     */
    public HashSet get_Grid2DSquareCells() {
        return this._Grid2DSquareCells;
    }
    
    /**
     * Sets this._Grid2DSquareCells to _Grid2DSquareCells.
     */
    public void set_Grid2DSquareCells(
            HashSet _Grid2DSquareCells ) {
        this._Grid2DSquareCells = _Grid2DSquareCells;
    }
    
    /**
     * Sets this._Grid2DSquareCells to _Grid2DSquareCells
     *
     *
     * @param _Grid2DSquareCells The HashSet this._Grid2DSquareCells is set to.
     */
    protected void set_Grid2DSquareCells(
            HashSet _Grid2DSquareCells,
            boolean _HandleOutOfMemoryError ) {
        try {
            this._Grid2DSquareCells = _Grid2DSquareCells;
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                set_Grid2DSquareCells(
                        _Grid2DSquareCells,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @param _Grid2DSquareCells A HashSet used for initialisation.
     */
    public void init_Grid2DSquareCells_MemoryReserve(
            HashSet _Grid2DSquareCells,
            boolean _HandleOutOfMemoryError ) {
        try {
            init_Grid2DSquareCells_MemoryReserve( _Grid2DSquareCells );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_Grid2DSquareCells_MemoryReserve(
                        _Grid2DSquareCells,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @param _Grid2DSquareCells A HashSet used for initialisation.
     */
    protected void init_Grid2DSquareCells_MemoryReserve(
            HashSet _Grid2DSquareCells ) {
        init_Grid2DSquareCells( _Grid2DSquareCells );
        Iterator _Iterator = this._Grid2DSquareCells.iterator();
        if ( _Iterator.hasNext() ) {
            AbstractGrid2DSquareCell _Grid2DSquareCell = ( AbstractGrid2DSquareCell ) _Iterator.next();
            this._MemoryReserve = _Grid2DSquareCell._MemoryReserve;
        } else {
            init_MemoryReserve();
        }
        //System.err.println( this.getClass().getName() + ".init_Grid2DSquareCells_MemoryReserve(HashSet)" );
    }
    
    /**
     * Initialises _MemoryReserve as an array of size given by size.
     *
     * @param size Size that _MemoryReserve is initialised to.
     */
    private final void init_MemoryReserve(
            int size ) {
        if ( this._MemoryReserve == null ) {
            this._MemoryReserve = new int[ size ];
            Arrays.fill(
                    this._MemoryReserve,
                    Integer.MIN_VALUE );
        }
    }
    
    /**
     * @return this._MemoryReserve.
     */
    public int[] get_MemoryReserve() {
        return this._MemoryReserve;
    }
    
    /**
     * Initialises _MemoryReserve.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void init_MemoryReserve(
            boolean _HandleOutOfMemoryError ) {
        try {
            init_MemoryReserve();
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Initialises _MemoryReserve. If an OutOfMemoryError is encountered
     * this calls _SwapToFileGrid2DSquareCellChunksExcept(_Grid2DSquareCell)
     * (if that returns null then it calls _SwapToFileGrid2DSquareCellChunk())
     * then recurses.
     * @param _Grid2DSquareCell
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void init_MemoryReserve(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            boolean _HandleOutOfMemoryError ) {
        try {
            init_MemoryReserve();
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                if ( _SwapToFileGrid2DSquareCellChunksExcept( _Grid2DSquareCell ) == null ) {
                     _SwapToFileGrid2DSquareCellChunk();
                }
//                _SwapToFileGrid2DSquareCellChunkExcept(
//                        _Grid2DSquareCell );
                init_MemoryReserve(
                        _Grid2DSquareCell,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Initialises _MemoryReserve. If an OutOfMemoryError is encountered
     * this calls
     * _SwapToFileGrid2DSquareCellChunksExcept(_ChunkID,_HandleOutOfMemoryError)
     * then recurses.
     * @param _ChunkID
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void init_MemoryReserve(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            init_MemoryReserve();
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                _SwapToFileGrid2DSquareCellChunksExcept(
                        _ChunkID );
//                _SwapToFileGrid2DSquareCellChunkExcept(
//                        _ChunkID );
                init_MemoryReserve(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Initialises _MemoryReserve. If an OutOfMemoryError is encountered
     * this calls
     * _SwapToFileGrid2DSquareCellChunksExcept(_Grid2DSquareCell,_ChunkID,_HandleOutOfMemoryError)
     * then recurses.
     * @param _Grid2DSquareCell
     * @param _ChunkID
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void init_MemoryReserve(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            init_MemoryReserve();
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                if ( _HandleOutOfMemoryError ) {
                    HashMap _HashMap = _SwapToFileGrid2DSquareCellChunksExcept(
                            _Grid2DSquareCell,
                            _ChunkID );
//                    HashMap _HashMap = _SwapToFileGrid2DSquareCellChunkExcept(
//                            _Grid2DSquareCell,
//                            _ChunkID );
                    if ( _HashMap == null ) {
                        System.out.println("DEBUG init_MemoryReserve(_Grid2DSquareCell,ChunkID,boolean");
                    }
                }
//                _SwapToFileGrid2DSquareCellChunkExcept(
//                        _Grid2DSquareCell,
//                        _ChunkID );
                init_MemoryReserve(
                        _Grid2DSquareCell,
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Initialises _MemoryReserve. If an OutOfMemoryError is encountered
     * this calls
     * _SwapToFileGrid2DSquareCellChunksExcept(_Grid2DSquareCell,_ChunkIDs,_HandleOutOfMemoryError)
     * then recurses.
     * @param _Grid2DSquareCell
     * @param _ChunkIDs HashSet of AbstractGrid2DSquareCell.ChunkIDs
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void init_MemoryReserve(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            HashSet _ChunkIDs,
            boolean _HandleOutOfMemoryError ) {
        try {
            init_MemoryReserve();
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                _SwapToFileGrid2DSquareCellChunksExcept(
                        _Grid2DSquareCell,
                        _ChunkIDs );
//                _SwapToFileGrid2DSquareCellChunkExcept(
//                        _Grid2DSquareCell,
//                        _ChunkIDs );
                init_MemoryReserve(
                        _Grid2DSquareCell,
                        _ChunkIDs,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Initialises _MemoryReserve. If an OutOfMemoryError is encountered
     * this calls
     * _SwapToFileGrid2DSquareCellChunksExcept(_Grid2DSquareCell_ChunkIDHashSet,_HandleOutOfMemoryError)
     * then recurses.
     * @param _Grid2DSquareCell_ChunkIDHashSet_HashMap HashMap with AbstractGrid2DSquareCell as
     * keys and a respective HashSet of AbstractGrid2DSquareCell.ChunkIDs. Identifying
     * those ChunkIDs not to be swapped from the AbstractGrid2DSquareCell.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void init_MemoryReserve(
            HashMap _Grid2DSquareCell_ChunkIDHashSet_HashMap,
            boolean _HandleOutOfMemoryError ) {
        try {
            init_MemoryReserve();
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                _SwapToFileGrid2DSquareCellChunksExcept(
                        _Grid2DSquareCell_ChunkIDHashSet_HashMap );
//                _SwapToFileGrid2DSquareCellChunkExcept(
//                        _Grid2DSquareCell_ChunkIDHashSet );
                init_MemoryReserve(
                        _Grid2DSquareCell_ChunkIDHashSet_HashMap,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Initialises _MemoryReserve as an array of size 4000000.
     * This is the default.
     * It was known that an array of size more 1000000 was needed up to 
     * Grids_1.0Beta8 for some test cases. In Grids_1.0Beta9 this was upped 
     * first to 2000000, then to 4000000 as lack of memory in swapping was 
     * thought to be causing a fault.
     * This memory reserved is used when an OutOfMemoryError is thrown. The 
     * memory reserve is cleared to allow data to be swapped from one location 
     * (memory/RAM) to another location (disc). The swapping may involve making 
     * a record of what data is swapped and returning this information.
     */
    private final void init_MemoryReserve() {
        init_MemoryReserve( 4000000 );
    }
    
    /**
     * Clears _MemoryReserve by setting it to null and calling the garbage
     * collector.
     * After this method is executed then no further method calls will
     * handle OutOfMemoryError until _MemoryReserve is initialsed.
     * It is usually assumed that when this method is called,
     * _MemoryReserve has been initialised apriori.
     */
    public final void clear_MemoryReserve() {
        //log(Runtime.getRuntime().freeMemory());
        this._MemoryReserve = null;
        System.gc();
    }
    
    /**
     * This allocates and deallocates a small amount of memory. It is used
     * to ensure there is enough memory upon stepping out to a method external
     * to Grids. What is needed at least is sufficient memory to step back
     * the Grids methods that handle OutOfMemoryErrors.
     * All the public methods of the API call this method before returning to
     * the rest of the program execution.
     * It exists for robustness reasons and was retrofitted as a work around to
     * the following problem:
     * What to do if upon returning there is insufficient memory to call another
     * method from the API that handles OutOfMemoryErrors.
     */
    public static void _EnsureThereIsEnoughMemoryToContinue() {
        try {
            // Initialise a small amount of memory
            int _intArraySize = 64;//512;//64;//1024;
            double[] _doubleArray0 = new double[ _intArraySize ];
            Arrays.fill( _doubleArray0, Double.MAX_VALUE );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            throw _OutOfMemoryError;
        }
    }
    
    /**
     * For returning the total free memory
     * <ul>
     * <li>( _Runtime.freeMemory() + ( _Runtime.maxMemory() - _Runtime.totalMemory() ) )</li>
     * </ul>
     * Even if this is zero then the JVM may be able to allocate memory in reserves?
     * @return The TotalFreeMemory available as calculated from _Runtime.
     * If the returned value is lower than a minimum required to call a method then probably best to swap some data.
     */
    protected long getTotalFreeMemory() {
        long result;
        try {
            long maxMemory = _Runtime.maxMemory();
            long allocatedMemory = _Runtime.totalMemory();
            long freeMemory = _Runtime.freeMemory();
            result = freeMemory + ( maxMemory - allocatedMemory );
            return result;
        } catch ( NullPointerException _NullPointerException ) {
            if ( _Runtime == null ) {
                init_Runtime();
            }
            return getTotalFreeMemory();
        }
    }
    
    /**
     * For returning the total free memory
     * <ul>
     * <li>( _Runtime.freeMemory() + ( _Runtime.maxMemory() - _Runtime.totalMemory() ) )</li>
     * </ul>
     * Even if this is zero then the JVM may be able to allocate memory in reserves?
     * @return The TotalFreeMemory available as calculated from _Runtime.
     * If the returned value is lower than a minimum required to call a method then probably best to swap some data.
     */
    public long getTotalFreeMemory(
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = getTotalFreeMemory();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getTotalFreeMemory(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to swap all AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught in this method then swap
     *     operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void _SwapToFileGrid2DSquareCellChunks(
            boolean _HandleOutOfMemoryError ) {
        try {
            _SwapToFileGrid2DSquareCellChunks();
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunks();
                init_MemoryReserve( _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to swap all AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells.
     */
    protected void _SwapToFileGrid2DSquareCellChunks() {
        Iterator _Iterator = this._Grid2DSquareCells.iterator();
        while ( _Iterator.hasNext() ) {
            ( ( AbstractGrid2DSquareCell ) _Iterator.next() ).swapToFileGrid2DSquareCellChunks( _HandleOutOfMemoryErrorFalse );
        }
    }
    
    /**
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells.
     * This is the lowest level of OutOfMemoryError handling in this class.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught in this method then swap
     *     operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunk(
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunk();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunk() {
        ChunkID _ChunkID = null;
        HashMap result = new HashMap();
        Iterator _Iterator = this._Grid2DSquareCells.iterator();
        while ( _Iterator.hasNext() ) {
            AbstractGrid2DSquareCell _Grid2DSquareCell =
                    ( AbstractGrid2DSquareCell ) _Iterator.next();
            _ChunkID = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunk(
                    _HandleOutOfMemoryErrorFalse );
            if ( _ChunkID != null ) {
                result.put( _Grid2DSquareCell, _ChunkID );
                return result;
            }
        }
        return null;
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for those in with AbstractGrid2DSquareCell.ChunkID = _ChunkID.
     * @param _ChunkID The AbstractGrid2DSquareCell.ChunkID not to be swapped.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                    _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                        _ChunkID );
                init_MemoryReserve(
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for those in with AbstractGrid2DSquareCell.ChunkID = _ChunkID.
     * @param _ChunkID The AbstractGrid2DSquareCell.ChunkID not to be swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            ChunkID _ChunkID ) {
        HashMap result = new HashMap();
        Iterator _Iterator = this._Grid2DSquareCells.iterator();
        while ( _Iterator.hasNext() ) {
            AbstractGrid2DSquareCell _Grid2DSquareCell = ( AbstractGrid2DSquareCell ) _Iterator.next();
            ChunkID _ChunkID0 = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunkExcept(
                    _ChunkID,
                    _HandleOutOfMemoryErrorFalse );
            if ( _ChunkID0 != null ) {
                result.put( _Grid2DSquareCell, new HashSet().add(_ChunkID0 ) );
                return result;
            }
        }
        return null;
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for those in _Grid2DSquareCell_ChunkIDHashSet.
     * @param _Grid2DSquareCell_ChunkIDHashSet HashMap with AbstractGrid2DSquareCell as
     * keys and a respective HashSet of AbstractGrid2DSquareCell.ChunkIDs. Identifying
     * those ChunkIDs not to be swapped from the AbstractGrid2DSquareCell.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            HashMap _Grid2DSquareCell_ChunkIDHashSet,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                    _Grid2DSquareCell_ChunkIDHashSet );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                        _Grid2DSquareCell_ChunkIDHashSet );
                init_MemoryReserve(
                        _Grid2DSquareCell_ChunkIDHashSet,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for those in _Grid2DSquareCell_ChunkIDHashSet.
     * @param _Grid2DSquareCell_ChunkIDHashSet HashMap with AbstractGrid2DSquareCell as
     * keys and a respective HashSet of AbstractGrid2DSquareCell.ChunkIDs. Identifying
     * those ChunkIDs not to be swapped from the AbstractGrid2DSquareCell.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            HashMap _Grid2DSquareCell_ChunkIDHashSet ) {
        HashMap result = new HashMap();
        Iterator _Iterator = this._Grid2DSquareCells.iterator();
        while ( _Iterator.hasNext() ) {
            AbstractGrid2DSquareCell _Grid2DSquareCell = ( AbstractGrid2DSquareCell ) _Iterator.next();
            if ( _Grid2DSquareCell_ChunkIDHashSet.containsKey( _Grid2DSquareCell ) ) {
                HashSet _ChunkIDs = ( HashSet ) _Grid2DSquareCell_ChunkIDHashSet.get( _Grid2DSquareCell );
                ChunkID _ChunkID = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunkExcept(
                        _ChunkIDs,
                        _HandleOutOfMemoryErrorFalse );
                if ( _ChunkID != null ) {
                    HashSet _SwappedChunkIDs = new HashSet();
                    _SwappedChunkIDs.add( _ChunkID );
                    result.put( _Grid2DSquareCell, _SwappedChunkIDs );
                    return result;
                }
            }
            ChunkID _ChunkID = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunk(
                    _HandleOutOfMemoryErrorFalse );
            if ( _ChunkID != null ) {
                HashSet _SwappedChunkIDs = new HashSet();
                _SwappedChunkIDs.add( _ChunkID );
                result.put( _Grid2DSquareCell, _SwappedChunkIDs );
                return result;
            }
        }
        return result; // If we get here then nothing could be swapped.
        //return null;
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for those in _Grid2DSquareCell with AbstractGrid2DSquareCell._ChunkID
     * in _ChunkIDs.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            HashSet _ChunkIDHashSet,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                    _Grid2DSquareCell,
                    _ChunkIDHashSet );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                        _Grid2DSquareCell,
                        _ChunkIDHashSet );
                init_MemoryReserve(
                        _Grid2DSquareCell,
                        _ChunkIDHashSet,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for those in _Grid2DSquareCell with AbstractGrid2DSquareCell._ChunkID
     * in _ChunkIDs.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            HashSet _ChunkIDs ) {
        HashMap result = new HashMap();
        ChunkID _ChunkID = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunkExcept(
                _ChunkIDs,
                _HandleOutOfMemoryErrorFalse );
        if ( _ChunkID != null ) {
            HashSet _SwappedChunkIDs = new HashSet();
            _SwappedChunkIDs.add( _ChunkID );
            result.put( _Grid2DSquareCell, _SwappedChunkIDs );
            return result;
        } else {
            return _SwapToFileGrid2DSquareCellChunkExcept(
                    _Grid2DSquareCell );
        }
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for that in _Grid2DSquareCell with AbstractGrid2DSquareCell._ChunkID
     * _ChunkID.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     * @param _ChunkID The AbstractGrid2DSquareCell.ChunkID not to be swapped.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                    _Grid2DSquareCell,
                    _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                        _Grid2DSquareCell,
                        _ChunkID );
                init_MemoryReserve(
                        _Grid2DSquareCell,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for that in _Grid2DSquareCell with AbstractGrid2DSquareCell._ChunkID
     * _ChunkID.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     * @param _ChunkID The AbstractGrid2DSquareCell.ChunkID not to be swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            ChunkID _ChunkID ) {
        HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                _Grid2DSquareCell );
        if ( result == null ) {
            result = new HashMap();
            HashSet _SwappedChunkIDs = new HashSet();
            ChunkID _ChunkIDSwapped = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunkExcept(
                    _ChunkID,
                    _HandleOutOfMemoryErrorFalse );
            if ( _ChunkIDSwapped == null ) {
                //System.out.println("Unable to swap in _SwapToFileGrid2DSquareCellChunkExcept_Protected(AbstractGrid2DSquareCell,ChunkID,boolean)");
                return null;
            } else {
                _SwappedChunkIDs.add( _ChunkIDSwapped );
                result.put( _Grid2DSquareCell, _SwappedChunkIDs );
                return result;
            }
        } else {
            return result;
        }
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for those in _Grid2DSquareCell.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                    _Grid2DSquareCell );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunkExcept(
                        _Grid2DSquareCell );
                init_MemoryReserve(
                        _Grid2DSquareCell,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * Attempts to swap any AbstractGrid2DSquareCellChunk in this._Grid2DSquareCells
     * except for those in _Grid2DSquareCell.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunkExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell ) {
        HashMap result = new HashMap();
        HashSet _Grid2DSquareCells_Grid2DSquareCell_Removed = new HashSet();
        _Grid2DSquareCells_Grid2DSquareCell_Removed.addAll( this._Grid2DSquareCells );
        _Grid2DSquareCells_Grid2DSquareCell_Removed.remove( _Grid2DSquareCell );
        Iterator _Iterator = _Grid2DSquareCells_Grid2DSquareCell_Removed.iterator();
        while( _Iterator.hasNext() ) {
            AbstractGrid2DSquareCell _Grid2DSquareCellToTryAndSwapFrom = ( AbstractGrid2DSquareCell ) _Iterator.next();
            ChunkID _ChunkID = _Grid2DSquareCellToTryAndSwapFrom.swapToFileGrid2DSquareCellChunk(
                    _HandleOutOfMemoryErrorFalse );
            if ( _ChunkID != null ) {
                HashSet _SwappedChunkIDs = new HashSet();
                _SwappedChunkIDs.add( _ChunkID );
                result.put( _Grid2DSquareCellToTryAndSwapFrom, _SwappedChunkIDs );
                return result;
            }
        }
        return null; // If we get here then nothing could be swapped.
    }
    
    /**
     * Attempts to Swap all AbstractGrid2DSquareCell.ChunkIDs in
     * this._Grid2DSquareCells except those with AbstractGrid2DSquareCell.ChunkID
     * _ChunkID.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * @param _ChunkID The AbstractGrid2DSquareCell.ChunkID not to be
     * swapped.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                    _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                        _ChunkID,
                        _HandleOutOfMemoryErrorFalse );
                init_MemoryReserve(
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to Swap all AbstractGrid2DSquareCell.ChunkIDs in
     * this._Grid2DSquareCells except those with AbstractGrid2DSquareCell.ChunkID
     * _ChunkID.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * @param _ChunkID The AbstractGrid2DSquareCell.ChunkID not to be
     * swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            ChunkID _ChunkID ) {
        HashMap result = new HashMap();
        Iterator _Iterator = this._Grid2DSquareCells.iterator();
        while ( _Iterator.hasNext() ) {
            AbstractGrid2DSquareCell _Grid2DSquareCell = ( AbstractGrid2DSquareCell ) _Iterator.next();
            result.put( _Grid2DSquareCell, _Grid2DSquareCell.swapToFileGrid2DSquareCellChunksExcept(
                    _ChunkID,
                    _HandleOutOfMemoryErrorFalse ) );
        }
        return result;
    }
    
    /**
     * Attempts to Swap all AbstractGrid2DSquareCell.ChunkIDs in
     * this._Grid2DSquareCells except those with AbstractGrid2DSquareCell.ChunkID
     * _ChunkID.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     * swapped.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                    _Grid2DSquareCell );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                        _Grid2DSquareCell,
                        _HandleOutOfMemoryErrorFalse );
                init_MemoryReserve(
                        _Grid2DSquareCell,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to Swap all AbstractGrid2DSquareCell.ChunkIDs in
     * this._Grid2DSquareCells except those with AbstractGrid2DSquareCell.ChunkID
     * _ChunkID.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell ) {
        HashMap result = new HashMap();
        HashSet _Grid2DSquareCells_Grid2DSquareCell_Removed = new HashSet();
        _Grid2DSquareCells_Grid2DSquareCell_Removed.addAll( this._Grid2DSquareCells );
        _Grid2DSquareCells_Grid2DSquareCell_Removed.remove( _Grid2DSquareCell );
        Iterator _Iterator = _Grid2DSquareCells_Grid2DSquareCell_Removed.iterator();
        while( _Iterator.hasNext() ) {
            AbstractGrid2DSquareCell _Grid2DSquareCellToTryAndSwapFrom = ( AbstractGrid2DSquareCell ) _Iterator.next();
            HashSet _SwappedChunkIDs = _Grid2DSquareCellToTryAndSwapFrom.swapToFileGrid2DSquareCellChunks(
                    _HandleOutOfMemoryErrorFalse );
            if ( _SwappedChunkIDs != null ) {
                result.put( _Grid2DSquareCellToTryAndSwapFrom, _SwappedChunkIDs );
            }
        }
        if ( result.isEmpty() ) {
            return null;
        }
        return result;
    }
    
    /**
     * Attempts to Swap all AbstractGrid2DSquareCell.ChunkIDs in
     * this._Grid2DSquareCells except those with AbstractGrid2DSquareCell.ChunkID
     * _ChunkID.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     * @param _ChunkID The AbstractGrid2DSquareCell.ChunkID not to be
     * swapped.
     */
    public HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                    _Grid2DSquareCell,
                    _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                        _ChunkID,
                        _HandleOutOfMemoryErrorFalse );
                init_MemoryReserve(
                        _Grid2DSquareCell,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to Swap all AbstractGrid2DSquareCell.ChunkIDs in
     * this._Grid2DSquareCells except those with AbstractGrid2DSquareCell.ChunkID
     * _ChunkID.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     * @param _ChunkID The AbstractGrid2DSquareCell.ChunkID not to be
     * swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            ChunkID _ChunkID ) {
        HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                _Grid2DSquareCell );
        if ( result == null ) {
            result = new HashMap();
        }
        HashSet _CellIDsSwappedFrom_Grid2DSquareCell = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunksExcept(
                _ChunkID,
                _HandleOutOfMemoryErrorFalse );
        if ( _CellIDsSwappedFrom_Grid2DSquareCell != null ) {
            result.put( _Grid2DSquareCell, _CellIDsSwappedFrom_Grid2DSquareCell );
        }
        if ( result.isEmpty() ) {
            return null;
        }
        return result;
    }
    
    /**
     * Attempts to Swap all AbstractGrid2DSquareCell.ChunkIDs in
     * this._Grid2DSquareCells except those with AbstractGrid2DSquareCell.ChunkIDs
     * in _ChunkIDs.
     * @return HashMap with:
     * key as the AbstractGrid2DSquareCell from which
     * the AbstractGrid2DSquareCellChunk was swapped; and,
     * value as the AbstractGrid2DSquareCellChunk._ChunkID swapped.
     * @param _Grid2DSquareCell AbstractGrid2DSquareCell that's chunks are not to be swapped.
     * @param _ChunkIDs The AbstractGrid2DSquareCell.ChunkID not to be
     * swapped.
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            HashSet _ChunkIDs ) {
        HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                _Grid2DSquareCell );
        if ( result == null ) {
            result = new HashMap();
        }
        HashSet _SwappedChunkIDs = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunksExcept(
                _ChunkIDs,
                _HandleOutOfMemoryErrorFalse );
        if ( _SwappedChunkIDs.size() > 0 ) {
            result.put( _Grid2DSquareCell, _SwappedChunkIDs );
        }
        if ( result.isEmpty() ) {
            return null;
        }
        return result;
    }
    
    /**
     */
    public HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            HashMap _Grid2DSquareCell_ChunkIDHashSet_HashMap,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                    _Grid2DSquareCell_ChunkIDHashSet_HashMap );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap result = _SwapToFileGrid2DSquareCellChunksExcept(
                        _Grid2DSquareCell_ChunkIDHashSet_HashMap );
                init_MemoryReserve(
                        _Grid2DSquareCell_ChunkIDHashSet_HashMap,
                        _HandleOutOfMemoryError );
                return result;
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     */
    protected HashMap _SwapToFileGrid2DSquareCellChunksExcept(
            HashMap _Grid2DSquareCell_ChunkIDHashSet_HashMap ) {
        HashMap result = new HashMap();
        Iterator _Iterator = this._Grid2DSquareCells.iterator();
        while ( _Iterator.hasNext() ) {
            AbstractGrid2DSquareCell _Grid2DSquareCell = ( AbstractGrid2DSquareCell ) _Iterator.next();
            if ( _Grid2DSquareCell_ChunkIDHashSet_HashMap.containsKey( _Grid2DSquareCell ) ) {
                HashSet _ChunkIDs = ( HashSet ) _Grid2DSquareCell_ChunkIDHashSet_HashMap.get( _Grid2DSquareCell );
                HashSet _SwappedChunkIDs = _Grid2DSquareCell.swapToFileGrid2DSquareCellChunksExcept(
                        _ChunkIDs,
                        _HandleOutOfMemoryErrorFalse );
                if ( _SwappedChunkIDs.size() > 0 ) {
                    result.put( _Grid2DSquareCell, _SwappedChunkIDs );
                }
            } else {
                result.put( _Grid2DSquareCell, _Grid2DSquareCell.swapToFileGrid2DSquareCellChunks(
                        _HandleOutOfMemoryErrorFalse ) );
            }
        }
        return result;
    }
    
    /**
     * For initialising a String of length length.
     */
    public String initString(
            int length,
            boolean _HandleOutOfMemoryError ) {
        try {
            String result = new String();
            for ( int i = 0; i < length; i ++ ) {
                result += "X";
            }
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return initString(
                        length,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return String which is _String_1 appended to the end of _String_0
     */
    public String initString(
            String _String_0,
            String _String_1,
            boolean _HandleOutOfMemoryError ) {
        try {
            String result = _String_0 + _String_1;
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return initString(
                        _String_0,
                        _String_1,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    
    /**
     * For initialising a File from String _String.
     */
    public File initFile(
            String _String,
            boolean _HandleOutOfMemoryError )
            throws IOException {
        try {
            File result = new File( _String );
            result.getParentFile().mkdirs();
            result.createNewFile();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return initFile(
                        _String,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * For initialising a File from File _File and String _String.
     */
    public File initFile(
            File _File,
            String _String,
            boolean _HandleOutOfMemoryError )
            throws IOException {
        try {
            File result = new File( _File, _String );
            _File.mkdirs();
            result.createNewFile();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return initFile(
                        _String,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * For initialising a File Directory from File _ParentFile and String _String.
     */
    public File initFileDirectory(
            File _ParentFile,
            String _String,
            boolean _HandleOutOfMemoryError )
            throws IOException {
        try {
            File result = new File( _ParentFile, _String );
            result.mkdirs();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return initFileDirectory(
                        _ParentFile,
                        _String,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    public String println(
            String _String,
            String _StringToDuplicateAndReturn,
            boolean _HandleOutOfMemoryError ) {
        try {
            System.out.println( _String );
            String result = new String( _StringToDuplicateAndReturn );
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return println(
                        _String,
                        _StringToDuplicateAndReturn,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    public double sin(
            double value,
            boolean _HandleOutOfMemoryError ) {
        try {
            return Math.sin( value );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return sin(
                        value,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    public String toString(
            double _double,
            boolean _HandleOutOfMemoryError ) {
        try {
            return Double.toString( _double );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return toString(
                        _double,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    public double[][] getNormalDistributionKernelWeights(
            AbstractGrid2DSquareCell _Grid2DSquareCell,
            double _Distance,
            boolean _HandleOutOfMemoryError ) {
        try {
            return uk.ac.leeds.ccg.andyt.grids.utilities.Kernel.getNormalDistributionKernelWeights(
                    _Grid2DSquareCell,
                    _Distance );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return getNormalDistributionKernelWeights(
                        _Grid2DSquareCell,
                        _Distance,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
}