/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.OutOfMemoryError;
import java.math.BigDecimal;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import uk.ac.leeds.ccg.andyt.grids.core.OutOfMemoryErrorHandler;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.UnsignedLongPowersOf2;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;
/**
 * Contains CellID and ChunkID classes, referencing and general geometry
 * methods. It also controls what methods extended classes must implement
 * acting like an interface.
 *
 * The basic geometries are ordered in set numbers of rows and columns and are
 * arranged sequentially as their base two-dimensional orthogonal coordinate
 * axes. The sequential arrangement goes along the x-axis row by row from the
 * y-axis, then up the y-axis taking each row in turn.
 *
 * TODO:
 * Handling for NumberFormatExceptions and ArithmeticExceptions in calculations
 */
public abstract class AbstractGrid2DSquareCell
        extends OutOfMemoryErrorHandler
        implements Serializable {
    
    /**
     * A version number for confidence in reloading serialised instances.
     */
    private static final long serialVersionUID = 1L;
    
    /**
     * Local _Directory used for caching.
     * TODO
     * This should not be transient so as it is possible to ascertain where
     * the data has come from.
     */
    protected transient File _Directory;
    
//    /** Abstracted to OutOfMemoryErrorHandler
//     * A collection of the Grid2dSquareCellAbstract Objects that
//     * may have data that can be swapped to release memory for
//     * processing.
//     */
//    public HashSet _Grid2DSquareCells;
    
    /**
     * The AbstractGrid2DSquareCellChunk data cache. A collection is used
     * rather than an array because an element of a collection set to null is
     * available for garbage collection whereas an element of an array set to
     * null seems not to be.
     */
    protected transient HashMap _Grid2DSquareCellChunks;
    
    /**
     * A reference to the grid Statistics Object.
     */
    protected AbstractGridStatistics _GridStatistics;
    
    /**
     * For storing the number of chunk rows.
     */
    protected int _NChunkRows;
    
    /**
     * For storing the number of chunk columns.
     */
    protected int _NChunkCols;
    
    /**
     * For storing the (usual) number of rows of cells in a chunk. The
     * number of rows in the final chunk row may be less.
     */
    protected int _ChunkNRows;
    
    /**
     * For storing the (usual) number of columns of cells in a chunk. The
     * number of columns in the final chunk column may be less.
     */
    protected int _ChunkNCols;
    
    /**
     * For storing the number of rows in the grid.
     */
    protected long _NRows;
    
    /**
     * For storing the number of columns in the grid.
     */
    protected long _NCols;
    
    /**
     * For storing the _Name of the grid.
     */
    protected String _Name;
    
    /**
     * For storing cellsize, minx, miny, maxx, maxy. Although maxx and maxy
     * could be easily calculated if required, for convenience they are
     * calculated and stored by default during construction.
     */
    protected BigDecimal[] _Dimensions;
    
    /**
     * For storing the minimum number of decimal places used to store
     * _Dimensions.
     */
    protected int _DimensionsScale;
    
    /**
     * For storing individual locations mapped to a binary encoded long. This
     * is only used in Grid2DSquareCellDoubleChunk64CellMap and
     * Grid2DSquareCellIntChunk64CellMap. It is stored in this to save it
     * being stored in every chunk or calculated on the fly.
     */
    protected static UnsignedLongPowersOf2 _UnsignedLongPowersOf2;
    
    /**
     * Initialises this._Grid2DSquareCellChunks by first attempting to load from
     * new File(
     *     gridFile,
     *     "cache" );
     * @param gridFile The _Directory that should may contain a File named cache.
     */
    protected void initGrid2DSquareCellChunks(
            File gridFile ) {
        try {
            File cache = new File(
                    gridFile,
                    "cache"  );
            if ( cache.exists() ) {
                try {
                    ObjectInputStream ois = new ObjectInputStream(
                            new BufferedInputStream(
                            new FileInputStream( cache ) ) );
                    this._Grid2DSquareCellChunks = ( HashMap ) ois.readObject();
                    ois.close();
                } catch ( Exception e0 ) {
                    System.err.println(
                            e0.getMessage() + " in \n" +
                            "initGrid2DSquareCellChunks( \n" +
                            "    File ( " + gridFile.toString() + " ) )" );
                }
            } else {
                this._Grid2DSquareCellChunks = new HashMap();
            }
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            System.out.println(
                    "Failed to load cache due to lack of memory" );
        }
    }
    
    /**
     * Initialises AbstractGrid2DSquareCell fields of this.
     * @param _Grid2DSquareCells HashSet
     */
    protected void initGrid2DSquareCell(
            HashSet _Grid2DSquareCells ) {
        init_Grid2DSquareCells_MemoryReserve( _Grid2DSquareCells );
        this._ChunkNCols = 1;
        this._ChunkNRows = 1;
        this._Dimensions = new BigDecimal[ 5 ];
        this._Dimensions[ 0 ] = new BigDecimal( 1L );
        this._Dimensions[ 1 ] = new BigDecimal( 0L );
        this._Dimensions[ 2 ] = new BigDecimal( 0L );
        this._Dimensions[ 3 ] = new BigDecimal( _ChunkNCols );
        this._Dimensions[ 4 ] = new BigDecimal( _ChunkNRows );
        this._DimensionsScale = 1;
        //this._Directory = FileCreator.createNewFile();
        this._Directory = new File(
                initString( 1000, _HandleOutOfMemoryErrorFalse ) );
        this._Grid2DSquareCellChunks = new HashMap();
        //this._Grid2DSquareCells = new HashSet();
        this._GridStatistics = new GridStatistics0();
        // Set the reference to this in the Grid Statistics
        this._GridStatistics.init( this );
        //this._GridStatistics.grid2DSquareCell = this;
        this._NChunkCols = 1;
        this._NCols = 1L;
        this._NRows = 1L;
        this._Name = new String(
                "DefaultName_" + getBasicDescription() );
        this._Grid2DSquareCells.add( this );
    }
    
    /**
     * Initialises non transient AbstractGrid2DSquareCell fields from
     * _Grid2DSquareCell.
     * @param _Grid2DSquareCell The AbstractGrid2DSquareCell from which the
     *   non transient AbstractGrid2DSquareCell fields of this are set.
     */
    protected void initGrid2DSquareCell(
            AbstractGrid2DSquareCell _Grid2DSquareCell ) {
        this._ChunkNCols = _Grid2DSquareCell._ChunkNCols;
        this._ChunkNRows = _Grid2DSquareCell._ChunkNRows;
        this._Dimensions = _Grid2DSquareCell._Dimensions;
        this._DimensionsScale = _Grid2DSquareCell._DimensionsScale;
        //this._Directory = _Grid2DSquareCell._Directory;
        this._GridStatistics = _Grid2DSquareCell._GridStatistics;
        // Set the reference to this in the Grid Statistics
        this._GridStatistics.init( this );
        //this._GridStatistics._Grid2DSquareCell = this;
        this._Name = _Grid2DSquareCell._Name;
        this._NChunkCols = _Grid2DSquareCell._NChunkCols;
        this._NChunkRows = _Grid2DSquareCell._NChunkRows;
        this._NCols = _Grid2DSquareCell._NCols;
        this._NRows = _Grid2DSquareCell._NRows;
        this._UnsignedLongPowersOf2 = _Grid2DSquareCell._UnsignedLongPowersOf2;
        //init_Grid2DSquareCells( _Grid2DSquareCell._Grid2DSquareCells );
        //this._Grid2DSquareCells = _Grid2DSquareCell._Grid2DSquareCells;
        this._Grid2DSquareCells.add( this );
    }
    
    /**
     * @return a reference to this._Grid2DSquareCells
     */
    public HashSet getGrid2DSquareCells(
            boolean _HandleOutOfMemoryError ) {
        try {
            HashSet result = get_Grid2DSquareCells();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getGrid2DSquareCells(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return this._GridStatistics
     * TODO:
     * For safety, this method should either be removed and this class be made
     * implement GridStatisticsInterface. This done the methods introduced
     * would be made to call the relevant ones in this._GridStatistics. Or
     * the _GridStatistics need to be made safe in that only copies of fields are
     * passed.
     */
    public AbstractGridStatistics getGridStatistics(
            boolean _HandleOutOfMemoryError ) {
        try {
            AbstractGridStatistics result = getGridStatistics();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getGridStatistics(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return this._GridStatistics
     * TODO:
     * For safety, this method should either be removed and this class be made
     * implement GridStatisticsInterface. This done the methods introduced
     * would be made to call the relevant ones in this._GridStatistics. Or
     * the _GridStatistics need to be made safe in that only copies of fields are
     * passed.
     */
    protected AbstractGridStatistics getGridStatistics( ) {
//        if ( this._GridStatistics.grid2DSquareCell != this ) {
//            boolean DEBUG = true;
//        }
        return this._GridStatistics;
    }
    
    /**
     * @return the AbstractGrid2DSquareCellChunk with ChunkID._ChunkRowIndex equal
     * to _ChunkRowIndex and ChunkID._ChunkColIndex equal to _ChunkColIndex.
     * @param _ChunkRowIndex The ChunkID._ChunkRowIndex of the returned
     *   AbstractGrid2DSquareCellChunk.
     * @param _ChunkColIndex The ChunkID._ChunkColIndex of the returned
     *   AbstractGrid2DSquareCellChunk.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected AbstractGrid2DSquareCellChunk getGrid2DSquareCellChunk(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            return getGrid2DSquareCellChunk(
                    _ChunkRowIndex,
                    _ChunkColIndex );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            _ChunkRowIndex,
                            _ChunkColIndex );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getGrid2DSquareCellChunk(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the AbstractGrid2DSquareCellChunk with ChunkID._ChunkRowIndex equal
     * to _ChunkRowIndex and ChunkID._ChunkColIndex equal to _ChunkColIndex.
     * @param _ChunkRowIndex The ChunkID._ChunkRowIndex of the returned
     *   AbstractGrid2DSquareCellChunk.
     * @param _ChunkColIndex The ChunkID._ChunkColIndex of the returned
     *   AbstractGrid2DSquareCellChunk.
     */
    protected AbstractGrid2DSquareCellChunk getGrid2DSquareCellChunk(
            int _ChunkRowIndex,
            int _ChunkColIndex ) {
        ChunkID _ChunkID = new ChunkID(
                this._NChunkCols,
                _ChunkRowIndex,
                _ChunkColIndex );
        return getGrid2DSquareCellChunk( _ChunkID );
    }
    
    /**
     * @return the AbstractGrid2DSquareCellChunk with ChunkID equal to _ChunkID.
     * @param _ChunkID The ChunkID of the returned AbstractGrid2DSquareCellChunk.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public AbstractGrid2DSquareCellChunk getGrid2DSquareCellChunk(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            AbstractGrid2DSquareCellChunk result =
                    getGrid2DSquareCellChunk( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunkExcept(
                        this,
                        _ChunkID );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return getGrid2DSquareCellChunk(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the AbstractGrid2DSquareCellChunk with ChunkID equal to _ChunkID.
     *
     *
     * @param _ChunkID The ChunkID of the returned AbstractGrid2DSquareCellChunk.
     */
    protected AbstractGrid2DSquareCellChunk getGrid2DSquareCellChunk(
            ChunkID _ChunkID ) {
        boolean isInGrid = isInGrid( _ChunkID );
        if ( isInGrid ) {
            boolean grid2DSquareCellChunksContainsKey =
                    this._Grid2DSquareCellChunks.containsKey( _ChunkID );
            if ( ! grid2DSquareCellChunksContainsKey ) {
                loadIntoCacheChunk(
                        _ChunkID );
            }
            return ( AbstractGrid2DSquareCellChunk ) this._Grid2DSquareCellChunks.get(
                    _ChunkID );
        }
        return null;
    }
    
    /**
     * @return HashSet containing all _Grid2DSquareCellChunks.ChunkID's.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public HashSet getGrid2DSquareCellChunkIDHashSet(
            boolean _HandleOutOfMemoryError ) {
        try {
            HashSet result = getGrid2DSquareCellChunkIDHashSet();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getGrid2DSquareCellChunkIDHashSet(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return HashSet containing all _Grid2DSquareCellChunks.ChunkID's.
     */
    protected HashSet getGrid2DSquareCellChunkIDHashSet() {
        HashSet result = new HashSet();
        result.addAll( this._Grid2DSquareCellChunks.keySet() );
        return result;
    }
    
    /**
     * @return Iterator over the cell values in this.
     */
    public abstract Iterator iterator(
            boolean _HandleOutOfMemoryError );
    
    /**
     * @return String description of this.
     */
    public abstract String toString(
            boolean _HandleOutOfMemoryError );
    
    /**
     * @return String description of this.
     * @param flag This is ignored. It is simply to distinguish this method from
     *   the abstract method toString(boolean).
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public String toString(
            int flag,
            boolean _HandleOutOfMemoryError ) {
        try {
            String result = toString( flag );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return toString(
                        flag,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a string description of the Abstract fields of this instance.
     * @param flag This is ignored. It is simply to distinguish this method from
     *   the abstract method toString(boolean).
     */
    protected String toString(
            int flag ) {
        boolean _HandleOutOfMemoryError = true;
        String result =
                "chunkNcols( " + this._ChunkNCols + " ), " +
                "chunkNrows( " + this._ChunkNRows + " ), " +
                "dimensionsScale( " + this._DimensionsScale + " ), " +
                "_NChunkCols( " + this._NChunkCols + " ), " +
                "_NChunkRows( " + this._NChunkRows + " ), " +
                "ncols( " + this._NCols + " ), " +
                "nrows( " + this._NRows + " ), " +
                "directory( " + this._Directory + " ), " +
                "name( " + this._Name + " ), " +
                "dimensions( " + this._Dimensions[ 0 ].toString();
        for ( int dimensionsID = 1; dimensionsID < this._Dimensions.length; dimensionsID ++ ) {
            result += ", " + _Dimensions[ dimensionsID ].toString();
        }
        if ( this._MemoryReserve == null ) {
            result += " ), memoryReserve( null";
        } else {
            result += " ), memoryReserve.length( " + this._MemoryReserve.length;
        }
        if ( this._Grid2DSquareCellChunks == null ) {
            result += " ), grid2DSquareCellChunks null";
        } else {
            result += " ), grid2DSquareCellChunks.size( " + this._Grid2DSquareCellChunks.size();
        }
        //result += " ), grid2DSquareCellIntProcessor( " + this.grid2DSquareCellIntProcessor.toString();
        if ( this._UnsignedLongPowersOf2 == null ) {
            result += " ), unsignedLongPowersOf2( null ), ";
        } else {
            result += " ), unsignedLongPowersOf2( " + this._UnsignedLongPowersOf2.toString();
        }
        result += " ), Statistics( " + this._GridStatistics.toString( _HandleOutOfMemoryError );
        if ( get_Grid2DSquareCells() == null ) {
            result += " ), grid2DSquareCells( null )";
        } else {
            result += " ), grid2DSquareCells.size( " + get_Grid2DSquareCells().size() + " )";
        }
        return result;
    }
    
    /**
     * @return a copy of _Directory.
     *
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public File get_Directory(
            boolean _HandleOutOfMemoryError ) {
        try {
            File result = get_Directory();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_Directory(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a copy of _Directory.
     */
    protected File get_Directory() {
        return new File( this._Directory.getPath() );
    }
    
    /**
     * @return a copy of this._Name.
     *
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public String get_Name(
            boolean _HandleOutOfMemoryError ) {
        try {
            String result = get_Name();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_Name(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a copy of this._Name.
     */
    protected String get_Name() {
        return new String( this._Name );
    }
    
    /**
     * Sets this._Name to _Name.
     *
     * @param _Name The String this._Name is set to.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void set_Name(
            String _Name,
            boolean _HandleOutOfMemoryError ) {
        try {
            setName( _Name );
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                set_Name(
                        _Name,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Sets this._Name to _Name.
     *
     * @param _Name The String this._Name is set to.
     */
    protected void setName( String _Name ) {
        this._Name = _Name;
    }
    
    /**
     * @return a basic description of this instance.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public String getBasicDescription(
            boolean _HandleOutOfMemoryError ) {
        try {
            //return getBasicDescription();
            String result = getBasicDescription();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getBasicDescription(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a basic description of this instance.
     */
    protected String getBasicDescription() {
        return new String( "className(" + this.getClass().getName() +
                ")_directory(" + this._Directory +
                ")_nrows(" + this._NRows +
                ")_ncols(" + this._NCols +
                ")_chunkNrows(" + this._ChunkNRows +
                ")_chunkNcols(" + this._ChunkNCols + ")" );
    }
    
    /**
     * Sets the AbstractGrid2DSquareCellChunk with ChunkID equal to _ChunkID to
     * _Grid2DSquareCellChunk.
     *
     *
     *
     * @param _Grid2DSquareCellChunk The AbstractGrid2DSquareCellChunk that the
     * AbstractGrid2DSquareCellChunk with ChunkID equal to _ChunkID is set as.
     * @param _ChunkID The ChunkID of the AbstractGrid2DSquareCellChunk that is set.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setChunk(
            AbstractGrid2DSquareCellChunk _Grid2DSquareCellChunk,
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            this._Grid2DSquareCellChunks.put(
                    _ChunkID,
                    _Grid2DSquareCellChunk );
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap _Grid2DSquareCell_ChunkIDHashSet = new HashMap();
                _Grid2DSquareCell_ChunkIDHashSet.put( this, _ChunkID );
                _SwapToFileGrid2DSquareCellChunkExcept(
                        _Grid2DSquareCell_ChunkIDHashSet );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                setChunk(
                        _Grid2DSquareCellChunk,
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a copy of this._NCols.
     *
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long get_NCols(
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = this._NCols;
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_NCols(
                        _HandleOutOfMemoryError);
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a copy of this._NRows.
     *
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long get_NRows(
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = this._NRows;
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_NRows(
                        _HandleOutOfMemoryError);
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a copy of this._NChunkRows.
     *
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int get_NChunkRows(
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = get_NChunkRows();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_NChunkRows(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a copy of this._NChunkRows.
     */
    public final int get_NChunkRows() {
        return this._NChunkRows;
    }
    
    /**
     * Initialises this._NChunkRows.
     */
    protected final void init_NChunkRows() {
        //        if ( ( this._NRows % this._ChunkNRows ) > 0 ) {
        //            this._NChunkRows = ( int ) (  this._NRows /  this._ChunkNRows ) + 1;
        //        } else {
        //            this._NChunkRows = ( int )  this._NRows /  this._ChunkNRows;
        //        }
        long longChunkNrows = ( long ) this._ChunkNRows;
        if ( ( this._NRows % longChunkNrows ) != 0 ) {
            this._NChunkRows = ( int ) ( this._NRows / longChunkNrows ) + 1;
        } else {
            this._NChunkRows = ( int ) ( this._NRows / longChunkNrows );
        }
    }
    
    /**
     * @return a copy of this._NChunkCols.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int get_NChunkCols(
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = get_NChunkCols();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_NChunkCols(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a copy of this._NChunkCols.
     */
    public final int get_NChunkCols(){
        return this._NChunkCols;
    }
    
    /**
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellDoubleChunkAbstracts in this as a long.
     */
    public final long getNChunks(
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = getNChunks();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getNChunks(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the number of Grid2DSquareCellDoubleChunkAbstracts in this as a long.
     */
    protected final long getNChunks() {
        long nChunks = ( long ) this._NChunkRows * ( long ) this._NChunkCols;
        if ( nChunks > Integer.MAX_VALUE ) {
            System.err.println(
                    "Error totalChunkCount > Integer.MAX_VALUE in \n" +
                    "initGrid2DSquareCellDouble( \n" +
                    "  GridStatisticsAbstract, \n" +
                    "  File, \n" +
                    "  Grid2DSquareCellDoubleChunkAbstractFactory, \n" +
                    "  int, \n" +
                    "  int, \n" +
                    "  long, \n" +
                    "  long, \n" +
                    "  BigDecimal[], \n" +
                    "  double )" );
            throw new Error();
        }
        return nChunks;
    }
    
    /**
     * Initialises this._NChunkCols.
     */
    protected final void init_NChunkCols() {
        //        if ( ( this._NCols % this._ChunkNCols ) > 0 ) {
        //            this._NChunkCols = ( int ) (  this._NCols /  this._ChunkNCols ) + 1;
        //        } else {
        //            this._NChunkCols = ( int )  this._NCols /  this._ChunkNCols;
        //        }
        long longChunkNcols = ( long ) this._ChunkNCols;
        if ( ( _NCols % longChunkNcols ) != 0 ) {
            this._NChunkCols = ( int ) ( this._NCols / longChunkNcols ) + 1;
        } else {
            this._NChunkCols = ( int ) ( this._NCols / longChunkNcols );
        }
    }
    
    /**
     * @return this._ChunkNRows.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int get_ChunkNRows(
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = get_ChunkNRows();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_ChunkNRows(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return this.getChunkNRows.
     */
    protected final int get_ChunkNRows() {
        return this._ChunkNRows;
    }
    /**
     * @return _ChunkNRows, the number of rows in AbstractGrid2DSquareCellChunk
     * with ChunkID._ChunkRowIndex equal to _ChunkRowIndex.
     * @param _ChunkRowIndex The ChunkID._ChunkRowIndex of the
     *   AbstractGrid2DSquareCellChunk thats _ChunkNRows is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int get_ChunkNRows(
            int _ChunkRowIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = get_ChunkNRows( _ChunkRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_ChunkNRows(
                        _ChunkRowIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return _ChunkNRows, the number of rows in AbstractGrid2DSquareCellChunk
     * with ChunkID._ChunkRowIndex equal to _ChunkRowIndex.
     * @param _ChunkRowIndex The ChunkID._ChunkRowIndex of the
     *   AbstractGrid2DSquareCellChunk thats _ChunkNRows is returned.
     */
    protected final int get_ChunkNRows(
            int _ChunkRowIndex ) {
        if ( _ChunkRowIndex > -1 && _ChunkRowIndex < this._NChunkRows ) {
            if ( _ChunkRowIndex == ( this._NChunkRows - 1 ) ) {
                return getChunkNrowsFinalRowChunks();
            } else {
                return this._ChunkNRows;
            }
        } else {
            return 0;
        }
    }
    
    /**
     * @return this._ChunkNCols.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int get_ChunkNCols(
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = get_ChunkNCols();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_ChunkNCols(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return this._ChunkNCols.
     */
    protected final int get_ChunkNCols() {
        return this._ChunkNCols;
    }
    
    /**
     * @return _ChunkNCols, the number of columns in AbstractGrid2DSquareCellChunk
     * with ChunkID._ChunkColIndex equal to _ChunkColIndex.
     * @param _ChunkColIndex The ChunkID._ChunkColIndex of the
     *   AbstractGrid2DSquareCellChunk thats _ChunkNRows is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int get_ChunkNCols(
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = get_ChunkNCols( _ChunkColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_ChunkNCols(
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return _ChunkNCols, the number of columns in AbstractGrid2DSquareCellChunk
     * with ChunkID._ChunkColIndex equal to _ChunkColIndex.
     * @param _ChunkColIndex The ChunkID._ChunkColIndex of the
     *   AbstractGrid2DSquareCellChunk thats _ChunkNRows is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param _ChunkID This is a ChunkID for those AbstractGrid2DSquareCells not
     * to be swapped if possible when an OutOfMemoryError is encountered.
     */
    public final int get_ChunkNCols(
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError,
            ChunkID _ChunkID ) {
        try {
            int result = get_ChunkNCols( _ChunkColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _SwapToFileGrid2DSquareCellChunkExcept( _ChunkID ) == null ) {
                    _SwapToFileGrid2DSquareCellChunk();
                }
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_ChunkNCols(
                        _ChunkColIndex,
                        _HandleOutOfMemoryError,
                        _ChunkID );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return _ChunkNCols, the number of columns in AbstractGrid2DSquareCellChunk
     * with ChunkID._ChunkColIndex equal to _ChunkColIndex.
     * @param _ChunkColIndex The ChunkID._ChunkColIndex of the
     *   AbstractGrid2DSquareCellChunk thats _ChunkNRows is returned.
     */
    protected final int get_ChunkNCols(
            int _ChunkColIndex ) {
        if ( _ChunkColIndex > -1 && _ChunkColIndex < this._NChunkCols ) {
            if ( _ChunkColIndex == ( this._NChunkCols - 1 ) ) {
                return getChunkNcolsFinalColChunks();
            } else {
                return this._ChunkNCols;
            }
        } else {
            return 0;
        }
    }
    
    /**
     * @return the number of rows in the final row Chunk.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final int getChunkNrowsFinalRowChunks(
            boolean _HandleOutOfMemoryError ) {
        try {
            return getChunkNrowsFinalRowChunks();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkNrowsFinalRowChunks(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the number of rows in the final row Chunk.
     */
    protected final int getChunkNrowsFinalRowChunks() {
        long longNChunkRowsMinusOne = ( long ) ( this._NChunkRows - 1 );
        long longChunkNrows = ( long ) this._ChunkNRows;
        return ( int ) ( this._NRows - ( longNChunkRowsMinusOne * longChunkNrows ) );
    }
    
    /**
     * @return the number of cols in the final col Chunk
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final int getChunkNcolsFinalColChunks(
            boolean _HandleOutOfMemoryError ) {
        try {
            return getChunkNcolsFinalColChunks();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkNcolsFinalColChunks(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the number of cols in the final col Chunk
     */
    protected final int getChunkNcolsFinalColChunks() {
        long longNChunkColsMinusOne = ( long ) ( this._NChunkCols - 1 );
        long longChunkNcols = ( long ) this._ChunkNCols;
        return ( int ) ( this._NCols - ( longNChunkColsMinusOne * longChunkNcols ) );
    }
    
    /**
     * @return _ChunkNRows, the number of rows in AbstractGrid2DSquareCellChunk with
     * ChunkID equal to _ChunkID
     * @param _ChunkID The ChunkID of the AbstractGrid2DSquareCellChunk thats number
     *   of rows is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNRows(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkNRows( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkNRows(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return _ChunkNRows, the number of rows in AbstractGrid2DSquareCellChunk with
     * ChunkID equal to _ChunkID
     * @param _ChunkID The ChunkID of the AbstractGrid2DSquareCellChunk thats number
     *   of rows is returned.
     */
    protected final int getChunkNRows(
            ChunkID _ChunkID ) {
        if ( _ChunkID._ChunkRowIndex < ( this._NChunkRows - 1 ) ) {
            return this._ChunkNRows;
        } else {
            return getChunkNrowsFinalRowChunks();
        }
    }
    
    /**
     * @return _ChunkNCols, the number of columns in AbstractGrid2DSquareCellChunk with
     * ChunkID equal to _ChunkID
     * @param _ChunkID The ChunkID of the AbstractGrid2DSquareCellChunk thats number
     *   of columns is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNCols(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkNCols( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkNCols(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return _ChunkNCols, the number of columns in AbstractGrid2DSquareCellChunk with
     * ChunkID equal to _ChunkID
     * @param _ChunkID The ChunkID of the AbstractGrid2DSquareCellChunk thats number
     *   of columns is returned.
     */
    protected final int getChunkNCols(
            ChunkID _ChunkID ) {
        if ( _ChunkID._ChunkColIndex < ( this._NChunkCols - 1 ) ) {
            return this._ChunkNCols;
        } else {
            return getChunkNcolsFinalColChunks();
        }
    }
    
    /**
     * @return A copy of this._Dimensions.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal[] get_Dimensions(
            boolean _HandleOutOfMemoryError ) {
        try {
            //double dimensions0 = this._Dimensions[0].doubleValue(); // For debugging
            //double dimensions1 = this._Dimensions[1].doubleValue(); // For debugging
            //double dimensions2 = this._Dimensions[2].doubleValue(); // For debugging
            //double dimensions3 = this._Dimensions[3].doubleValue(); // For debugging
            //double dimensions4 = this._Dimensions[4].doubleValue(); // For debugging
            int length = _Dimensions.length;
            BigDecimal[] result = new BigDecimal[ length ];
            System.arraycopy( _Dimensions, 0, result, 0, length );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_Dimensions(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return double equal to this._Dimensions[0].doubleValue.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellsizeDouble(
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = this._Dimensions[ 0 ].doubleValue();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellsizeDouble(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return this._DimensionsScale.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final int get_DimensionsScale(
            boolean _HandleOutOfMemoryError ) {
        try {
            return this._DimensionsScale;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return get_DimensionsScale(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Initialises this._Dimensions and this.dimensionScale.
     */
    protected final void init_Dimensions(
            BigDecimal[] _Dimensions ) {
        this._Dimensions = _Dimensions;
        this._DimensionsScale = Integer.MIN_VALUE;
        for ( int i = 0; i < _Dimensions.length; i ++ ) {
            this._DimensionsScale = Math.max(
                    this._DimensionsScale,
                    _Dimensions[ i ].scale() );
        }
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at at Point2D.Double point
     * as a double.
     * @param point The Point2D.Double for which the cell value is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            Point2D.Double point,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellDouble( point );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            getChunkRowIndex( point.getY() ),
                            getChunkColIndex( point.getX() ) );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getCellDouble(
                        point,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at at Point2D.Double point
     * as a double.
     * @param point The Point2D.Double for which the cell value is returned.
     */
    protected double getCellDouble(
            Point2D.Double point ) {
        return getCellDouble(
                getChunkRowIndex( point.getY() ),
                getChunkColIndex( point.getX() ),
                getChunkCellRowIndex( point.getY() ),
                getChunkCellColIndex( point.getX() ) );
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at at point given by
     * x-coordinate x and y-coordinate y as a double.
     *
     * @param x The x coordinate of the point at which the cell value is returned.
     * @param y The y coordinate of the point at which the cell value is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            double x,
            double y,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellDouble(
                    x,
                    y );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getCellDouble(
                        x,
                        y,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at at point given by
     * x-coordinate x and y-coordinate y as a double.
     * @param x The x coordinate of the point at which the cell value is returned.
     * @param y The y coordinate of the point at which the cell value is returned.
     */
    protected double getCellDouble(
            double x,
            double y ) {
        return getCellDouble(
                getChunkRowIndex( y ),
                getChunkColIndex( x ),
                getChunkCellRowIndex( y ),
                getChunkCellColIndex( x ) );
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at cell row index equal to
     * _CellRowIndex, cell col index equal to _CellColIndex as a double.
     * @param _CellRowIndex The cell row index of the .
     * @param _CellColIndex The cell column index.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            long _CellRowIndex,
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellDouble(
                    _CellRowIndex,
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            getChunkRowIndex( _CellRowIndex ),
                            getChunkColIndex( _CellColIndex ) );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getCellDouble(
                        _CellRowIndex,
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at cell row index equal to
     * _CellRowIndex, cell col index equal to _CellColIndex as a double.
     * @param _CellRowIndex The cell row index of the .
     * @param _CellColIndex The cell column index.
     */
    protected double getCellDouble(
            long _CellRowIndex,
            long _CellColIndex ) {
        return getCellDouble(
                getChunkRowIndex( _CellRowIndex ),
                getChunkColIndex( _CellColIndex ),
                getChunkCellRowIndex( _CellRowIndex ),
                getChunkCellColIndex( _CellColIndex ) );
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at cell row index equal to
     * _CellRowIndex, cell col index equal to _CellColIndex as a double.
     * @param _ChunkRowIndex The cell chunk row index.
     * @param _ChunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellDouble(
                    _ChunkRowIndex,
                    _ChunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            _ChunkRowIndex,
                            _ChunkColIndex );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getCellDouble(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at cell row index equal to
     * _CellRowIndex, cell col index equal to _CellColIndex as a double.
     * @param _ChunkRowIndex The cell chunk row index.
     * @param _ChunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     */
    protected double getCellDouble(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        if ( ! isInGrid( _ChunkRowIndex, _ChunkColIndex ) ) {
            if ( this instanceof Grid2DSquareCellDouble ) {
                return ( ( Grid2DSquareCellDouble ) this )._NoDataValue;
            } else {
                //this instanceof Grid2DSquareCellInt
                return ( ( Grid2DSquareCellInt ) this ).getNoDataValue();
            }
        }
        AbstractGrid2DSquareCellChunk grid2DSquareCellChunk =
                getGrid2DSquareCellChunk(
                _ChunkRowIndex,
                _ChunkColIndex );
        if ( grid2DSquareCellChunk == null ) {
            System.out.println("_ChunkRowIndex " + _ChunkRowIndex + ", _ChunkColIndex " + _ChunkColIndex );
            System.out.println("chunkCellRowIndex " + chunkCellRowIndex + ", chunkCellColIndex " + chunkCellColIndex );
            return -9999.0D;
            //return this.getNoDataValueBigDecimal( this._HandleOutOfMemoryErrorFalse ).doubleValue();
        }
        return getCellDouble(
                grid2DSquareCellChunk,
                _ChunkRowIndex,
                _ChunkColIndex,
                chunkCellRowIndex,
                chunkCellColIndex );
    }
    
    /**
     * @return Cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of AbstractGrid2DSquareCellChunk given by chunk
     * row index _ChunkRowIndex, chunk col index _ChunkColIndex as a double.
     * @param grid2DSquareCellChunk The AbstractGrid2DSquareCellChunk containing
     *   the cell.
     * @param _ChunkRowIndex The cell chunk row index.
     * @param _ChunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            AbstractGrid2DSquareCellChunk grid2DSquareCellChunk,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellDouble(
                    grid2DSquareCellChunk,
                    _ChunkRowIndex,
                    _ChunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            _ChunkRowIndex,
                            _ChunkColIndex );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getCellDouble(
                        grid2DSquareCellChunk,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of AbstractGrid2DSquareCellChunk given by chunk
     * row index _ChunkRowIndex, chunk col index _ChunkColIndex as a double.
     * @param grid2DSquareCellChunk The AbstractGrid2DSquareCellChunk containing
     *   the cell.
     * @param _ChunkRowIndex The cell chunk row index.
     * @param _ChunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     */
    protected double getCellDouble(
            AbstractGrid2DSquareCellChunk grid2DSquareCellChunk,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        if ( grid2DSquareCellChunk instanceof AbstractGrid2DSquareCellDoubleChunk ) {
            AbstractGrid2DSquareCellDoubleChunk grid2DSquareCellDoubleChunk =
                    ( AbstractGrid2DSquareCellDoubleChunk ) grid2DSquareCellChunk;
            Grid2DSquareCellDouble grid2DSquareCellDouble =
                    grid2DSquareCellDoubleChunk.getGrid2DSquareCellDouble();
            double noDataValue = grid2DSquareCellDouble._NoDataValue;
            if ( grid2DSquareCellChunk.getClass() ==
                    Grid2DSquareCellDoubleChunk64CellMap.class ) {
                Grid2DSquareCellDoubleChunk64CellMap grid2DSquareCellDoubleChunk64CellMap =
                        ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunk.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                Grid2DSquareCellDoubleChunkArray grid2DSquareCellDoubleChunkArray =
                        ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunkArray.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                Grid2DSquareCellDoubleChunkJAI grid2DSquareCellDoubleChunkJAI =
                        ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunkJAI.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                Grid2DSquareCellDoubleChunkMap grid2DSquareCellDoubleChunkMap =
                        ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunkMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkRAF.class ) {
                Grid2DSquareCellDoubleChunkRAF grid2DSquareCellDoubleChunkRAF =
                        ( Grid2DSquareCellDoubleChunkRAF ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunkRAF.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            return noDataValue;
        } else {
            //( grid2DSquareCellChunk instanceof AbstractGrid2DSquareCellIntChunk )
            AbstractGrid2DSquareCellIntChunk grid2DSquareCellIntChunk =
                    ( AbstractGrid2DSquareCellIntChunk ) grid2DSquareCellChunk;
            Grid2DSquareCellInt grid2DSquareCellInt =
                    grid2DSquareCellIntChunk.getGrid2DSquareCellInt();
            int noDataValue = grid2DSquareCellInt.getNoDataValue(
                    this._HandleOutOfMemoryErrorFalse );
            if ( grid2DSquareCellChunk.getClass() ==
                    Grid2DSquareCellIntChunk64CellMap.class ) {
                Grid2DSquareCellIntChunk64CellMap grid2DSquareCellIntChunk64CellMap =
                        ( Grid2DSquareCellIntChunk64CellMap ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunk.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                Grid2DSquareCellIntChunkArray grid2DSquareCellIntChunkArray =
                        ( Grid2DSquareCellIntChunkArray ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunkArray.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                Grid2DSquareCellIntChunkJAI grid2DSquareCellIntChunkJAI =
                        ( Grid2DSquareCellIntChunkJAI ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunkJAI.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                Grid2DSquareCellIntChunkMap grid2DSquareCellIntChunkMap =
                        ( Grid2DSquareCellIntChunkMap ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunkMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkRAF.class ) {
                Grid2DSquareCellIntChunkRAF grid2DSquareCellIntChunkRAF =
                        ( Grid2DSquareCellIntChunkRAF ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunkRAF.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            return ( double ) noDataValue;
        }
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at cell row index equal to
     * _CellRowIndex, cell col index equal to _CellColIndex as a int.
     * @param _CellRowIndex The cell row index.
     * @param _CellColIndex The cell column index.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt(
            long _CellRowIndex,
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getCellInt(
                    _CellRowIndex,
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            getChunkRowIndex( _CellRowIndex ),
                            getChunkColIndex( _CellColIndex ) );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getCellInt(
                        _CellRowIndex,
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return AbstractGrid2DSquareCellChunk cell value at cell row index equal to
     * _CellRowIndex, cell col index equal to _CellColIndex as a int.
     * @param _CellRowIndex The cell row index.
     * @param _CellColIndex The cell column index.
     */
    protected int getCellInt(
            long _CellRowIndex,
            long _CellColIndex ) {
        return getCellInt(
                getChunkRowIndex( _CellRowIndex ),
                getChunkColIndex( _CellColIndex ),
                getChunkCellRowIndex( _CellRowIndex ),
                getChunkCellColIndex( _CellColIndex ) );
    }
    
    /**
     * @return Cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of AbstractGrid2DSquareCellChunk given by chunk
     * row index _ChunkRowIndex, chunk col index _ChunkColIndex as a int.
     * @param _ChunkRowIndex The cell chunk row index.
     * @param _ChunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getCellInt(
                    _ChunkRowIndex,
                    _ChunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            _ChunkRowIndex,
                            _ChunkColIndex );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getCellInt(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of AbstractGrid2DSquareCellChunk given by chunk
     * row index _ChunkRowIndex, chunk col index _ChunkColIndex as a int.
     *
     *
     * @param _ChunkRowIndex The cell chunk row index.
     * @param _ChunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     */
    protected int getCellInt(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        AbstractGrid2DSquareCellChunk grid2DSquareCellChunk = getGrid2DSquareCellChunk(
                _ChunkRowIndex,
                _ChunkColIndex );
        return getCellInt(
                grid2DSquareCellChunk,
                _ChunkRowIndex,
                _ChunkColIndex,
                chunkCellRowIndex,
                chunkCellColIndex );
    }
    
    /**
     * @return Cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of AbstractGrid2DSquareCellChunk given by chunk
     * row index _ChunkRowIndex, chunk col index _ChunkColIndex as a int.
     * @param grid2DSquareCellChunk The AbstractGrid2DSquareCellChunk containing
     *   the cell.
     * @param _ChunkRowIndex The cell chunk row index.
     * @param _ChunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt(
            AbstractGrid2DSquareCellChunk grid2DSquareCellChunk,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getCellInt(
                    grid2DSquareCellChunk,
                    _ChunkRowIndex,
                    _ChunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            _ChunkRowIndex,
                            _ChunkColIndex );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return getCellInt(
                        grid2DSquareCellChunk,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of AbstractGrid2DSquareCellChunk given by chunk
     * row index _ChunkRowIndex, chunk col index _ChunkColIndex as a int.
     * @param grid2DSquareCellChunk The AbstractGrid2DSquareCellChunk containing
     *   the cell.
     * @param _ChunkRowIndex The cell chunk row index.
     * @param _ChunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     */
    protected int getCellInt(
            AbstractGrid2DSquareCellChunk grid2DSquareCellChunk,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        if ( grid2DSquareCellChunk instanceof AbstractGrid2DSquareCellDoubleChunk ) {
            AbstractGrid2DSquareCellDoubleChunk grid2DSquareCellDoubleChunk =
                    ( AbstractGrid2DSquareCellDoubleChunk ) grid2DSquareCellChunk;
            Grid2DSquareCellDouble grid2DSquareCellDouble =
                    grid2DSquareCellDoubleChunk.getGrid2DSquareCellDouble();
            double noDataValue = grid2DSquareCellDouble._NoDataValue;
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                Grid2DSquareCellDoubleChunk64CellMap grid2DSquareCellDoubleChunk64CellMap =
                        ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunk64CellMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                Grid2DSquareCellDoubleChunkArray grid2DSquareCellDoubleChunkArray =
                        ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunkArray.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                Grid2DSquareCellDoubleChunkJAI grid2DSquareCellDoubleChunkJAI =
                        ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunkJAI.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                Grid2DSquareCellDoubleChunkMap grid2DSquareCellDoubleChunkMap =
                        ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunkMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkRAF.class ) {
                Grid2DSquareCellDoubleChunkRAF grid2DSquareCellDoubleChunkRAF =
                        ( Grid2DSquareCellDoubleChunkRAF ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunkRAF.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            return ( int ) noDataValue;
        } else {
            //if ( grid2DSquareCellChunk instanceof AbstractGrid2DSquareCellIntChunk ) {
            AbstractGrid2DSquareCellIntChunk grid2DSquareCellIntChunk =
                    ( AbstractGrid2DSquareCellIntChunk ) grid2DSquareCellChunk;
            Grid2DSquareCellInt grid2DSquareCellInt =
                    grid2DSquareCellIntChunk.getGrid2DSquareCellInt();
            int noDataValue = grid2DSquareCellInt.getNoDataValue();
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunk64CellMap.class ) {
                Grid2DSquareCellIntChunk64CellMap grid2DSquareCellIntChunk64CellMap =
                        ( Grid2DSquareCellIntChunk64CellMap ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunk64CellMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                Grid2DSquareCellIntChunkArray grid2DSquareCellIntChunkArray =
                        ( Grid2DSquareCellIntChunkArray ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunkArray.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                Grid2DSquareCellIntChunkJAI grid2DSquareCellIntChunkJAI =
                        ( Grid2DSquareCellIntChunkJAI ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunkJAI.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                Grid2DSquareCellIntChunkMap grid2DSquareCellIntChunkMap =
                        ( Grid2DSquareCellIntChunkMap ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunkMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkRAF.class ) {
                Grid2DSquareCellIntChunkRAF grid2DSquareCellIntChunkRAF =
                        ( Grid2DSquareCellIntChunkRAF ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunkRAF.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            return noDataValue;
        }
    }
    
    /**
     * @return the value at _CellRowIndex, _CellColIndex as a double and sets it to valueToSet.
     * @param _CellRowIndex The cell row index.
     * @param _CellColIndex The cell column index.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double setCell(
            long _CellRowIndex,
            long _CellColIndex,
            double valueToSet,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = setCell(
                    _CellRowIndex,
                    _CellColIndex,
                    valueToSet );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            get_NChunkCols(),
                            getChunkRowIndex( _CellRowIndex ),
                            getChunkColIndex( _CellColIndex ) );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return setCell(
                        _CellRowIndex,
                        _CellColIndex,
                        valueToSet,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the value at _CellRowIndex, _CellColIndex as a double and sets it to valueToSet.
     * @param _CellRowIndex The cell row index.
     * @param _CellColIndex The cell column index.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param _ChunkID This is a ChunkID for those AbstractGrid2DSquareCells not
     * to be swapped if possible when an OutOfMemoryError is encountered.
     */
    public double setCell(
            long _CellRowIndex,
            long _CellColIndex,
            double valueToSet,
            boolean _HandleOutOfMemoryError,
            ChunkID _ChunkID ) {
        try {
            double result = setCell(
                    _CellRowIndex,
                    _CellColIndex,
                    valueToSet );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID2 = new ChunkID(
                            get_NChunkCols(),
                            getChunkRowIndex( _CellRowIndex ),
                            getChunkColIndex( _CellColIndex ) );
                    if ( _SwapToFileGrid2DSquareCellChunkExcept( _ChunkID2, _HandleOutOfMemoryError ) == null ) {
                        _SwapToFileGrid2DSquareCellChunk( _HandleOutOfMemoryError );
                    }
                    init_MemoryReserve(
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return setCell(
                        _CellRowIndex,
                        _CellColIndex,
                        valueToSet,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the value at _CellRowIndex, _CellColIndex as a double and sets it to valueToSet.
     * @param _CellRowIndex The cell row index.
     * @param _CellColIndex The cell column index.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param _Grid2DSquareCell This is a AbstractGrid2DSquareCell not
     * to be swapped if possible when an OutOfMemoryError is encountered.
     */
    public double setCell(
            long _CellRowIndex,
            long _CellColIndex,
            double valueToSet,
            boolean _HandleOutOfMemoryError,
            AbstractGrid2DSquareCell _Grid2DSquareCell ) {
        try {
            double result = setCell(
                    _CellRowIndex,
                    _CellColIndex,
                    valueToSet );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    if ( _SwapToFileGrid2DSquareCellChunkExcept( _Grid2DSquareCell, _HandleOutOfMemoryError ) == null ) {
                        _SwapToFileGrid2DSquareCellChunk( _HandleOutOfMemoryError );
                    }
                    init_MemoryReserve(
                            _Grid2DSquareCell,
                            _HandleOutOfMemoryError );
                }
                return setCell(
                        _CellRowIndex,
                        _CellColIndex,
                        valueToSet,
                        _HandleOutOfMemoryError,
                        _Grid2DSquareCell );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Sets the value at _CellRowIndex, _CellColIndex to valueToSet.
     * @param _CellRowIndex The cell row index.
     * @param _CellColIndex The cell column index.
     * @param valueToSet The value set.
     */
    protected abstract double setCell(
            long _CellRowIndex,
            long _CellColIndex,
            double valueToSet );
    
    
    /**
     * @return HashSet of ChunkIDs for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x The x-coordinate of the circle centre from which cell values are
     *   returned.
     * @param y The y-coordinate of the circle centre from which cell values are
     *   returned.
     * @param _CellRowIndex The row index at y.
     * @param _CellColIndex The col index at x.
     * @param distance The radius of the circle for which intersected cell
     *   values are returned.
     */
    protected HashSet getChunkIDs(
            double distance,
            double x,
            double y,
            long _CellRowIndex,
            long _CellColIndex ) {
        HashSet result = new HashSet();
        long p;
        long q;
        long cellDistance = ( long ) Math.ceil( distance / _Dimensions[ 0 ].doubleValue() );
        double thisCellX;
        double thisCellY;
        double thisDistance;
        for ( p = - cellDistance; p <= cellDistance; p ++ ) {
            thisCellY = getCellYDouble( _CellRowIndex + p );
            for ( q = - cellDistance; q <= cellDistance; q ++ ) {
                thisCellX = getCellXDouble( _CellColIndex + q );
                thisDistance = Utilities.distance( thisCellX, thisCellY, x, y );
                if ( thisDistance < distance ) {
                    ChunkID _ChunkID = new ChunkID(
                            get_ChunkNRows(),
                            getChunkRowIndex( ( long ) _CellRowIndex + p ),
                            getChunkColIndex( ( long ) _CellColIndex + q ) );
                    result.add( _ChunkID );
                }
            }
        }
        return result;
    }
    
    /**
     * @return Chunk column index for the AbstractGrid2DSquareCellChunk
     * intersecting the x-coordinate x.
     *
     * @param x The x-coordinate of the line intersecting the chunk column
     *   index returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkColIndex(
            double x,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkColIndex( x );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkColIndex(
                        x,
                        _HandleOutOfMemoryError);
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Chunk column index for the AbstractGrid2DSquareCellChunk
     * intersecting the x-coordinate x.
     * @param x The x-coordinate of the line intersecting the chunk column
     *   index returned.
     */
    protected final int getChunkColIndex(
            double x ) {
        return getChunkColIndex( getCellColIndex( x ) );
    }
    
    /**
     * @return Chunk column index for the AbstractGrid2DSquareCellChunk
     * intersecting the cell column index _CellColIndex.
     * @param _CellColIndex
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkColIndex(
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkColIndex( _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                return getChunkColIndex(
                        _CellColIndex,
                        _HandleOutOfMemoryError);
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Chunk column index for the AbstractGrid2DSquareCellChunk
     * intersecting the cell column index _CellColIndex.
     *
     * @param _CellColIndex
     */
    protected final int getChunkColIndex(
            long _CellColIndex ) {
        return ( int ) ( _CellColIndex / ( long ) this._ChunkNCols );
    }
    
    /**
     * @return Cell column Index for the cell column that intersect the
     * x-coordinate x.
     * @param x The x-coordinate of the line intersecting the cell column
     *   index returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellColIndex(
            double x,
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = getCellColIndex( x );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellColIndex(
                        x,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell column Index for the cell column that intersect the
     * x-coordinate x.
     * @param x The x-coordinate of the line intersecting the cell column
     *   index returned.
     */
    protected final long getCellColIndex(
            double x ) {
        return getCellColIndex( BigDecimal.valueOf( x ) );
    }
    
    /**
     * @return Cell column Index for the cell column that intersect the
     * x-coordinate xBigDecimal.
     * @param xBigDecimal The x-coordinate of the line intersecting the cell column
     *   index returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellColIndex(
            BigDecimal xBigDecimal,
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = getCellColIndex( xBigDecimal );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellColIndex(
                        xBigDecimal,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell column Index for the cell column that intersect the
     * x-coordinate xBigDecimal.
     * @param xBigDecimal The x-coordinate of the line intersecting the cell column
     *   index returned.
     */
    protected final long getCellColIndex(
            BigDecimal xBigDecimal ) {
        int scale = Math.max(
                xBigDecimal.scale(),
                this._DimensionsScale );
        BigDecimal _BigDecimal0 = xBigDecimal.subtract(
                this._Dimensions[ 1 ] );
        BigDecimal _BigDecimal1 = _BigDecimal0.divide(
                this._Dimensions[ 0 ],
                scale,
                BigDecimal.ROUND_DOWN );
        return _BigDecimal1.longValue();
    }
    
    /**
     * @return Cell column index for the cells in chunk column index
     * _ChunkColIndex chunk cell column index chunkCellColIndex.
     * @param _ChunkColIndex
     * @param chunkCellColIndex
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellColIndex(
            int _ChunkColIndex,
            int chunkCellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = getCellColIndex(
                    _ChunkColIndex,
                    chunkCellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellColIndex(
                        _ChunkColIndex,
                        chunkCellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell column index for the cells in chunk column index
     * _ChunkColIndex chunk cell column index chunkCellColIndex.
     * @param _ChunkColIndex
     * @param chunkCellColIndex
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param _ChunkID This is a ChunkID for those AbstractGrid2DSquareCells not
     * to be swapped if possible when an OutOfMemoryError is encountered.
     */
    public final long getCellColIndex(
            int _ChunkColIndex,
            int chunkCellColIndex,
            boolean _HandleOutOfMemoryError,
            ChunkID _ChunkID ) {
        try {
            long result = getCellColIndex(
                    _ChunkColIndex,
                    chunkCellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _SwapToFileGrid2DSquareCellChunkExcept( _ChunkID ) == null ) {
                    _SwapToFileGrid2DSquareCellChunk();
                }
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellColIndex(
                        _ChunkColIndex,
                        chunkCellColIndex,
                        _HandleOutOfMemoryError,
                        _ChunkID );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell column index for the cells in chunk column index
     * _ChunkColIndex chunk cell column index chunkCellColIndex.
     *
     * @param _ChunkColIndex
     * @param chunkCellColIndex
     */
    protected final long getCellColIndex(
            int _ChunkColIndex,
            int chunkCellColIndex ) {
        return ( ( long ) _ChunkColIndex * ( long ) this._ChunkNCols ) + ( long ) chunkCellColIndex;
    }
    
    /**
     * @return Chunk cell column Index of the cells that intersect the
     * x-coordinate x.
     * @param x The x-coordinate of the line intersecting the chunk cell column
     *   index returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellColIndex(
            double x,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkCellColIndex( x );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkCellColIndex(
                        x,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Chunk cell column Index of the cells that intersect the
     * x-coordinate x.
     * @param x The x-coordinate of the line intersecting the chunk cell column
     *   index returned.
     */
    protected final int getChunkCellColIndex(
            double x ) {
        return getChunkCellColIndex( getCellColIndex( x ) );
    }
    
    /**
     * @return Chunk cell column index of the cells in the cell column index
     * _CellColIndex.
     * @param _CellColIndex The cell column index of the cell thats chunk cell
     *   column index is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellColIndex(
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkCellColIndex( _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkCellColIndex(
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Chunk cell column index of the cells in the cell column index
     * _CellColIndex.
     * @param _CellColIndex The cell column index of the cell thats chunk cell
     *   column index is returned.
     */
    protected final int getChunkCellColIndex(
            long _CellColIndex ) {
        return ( int ) ( _CellColIndex - ( ( _CellColIndex / this._ChunkNCols ) * this._ChunkNCols ) );
    }
    
    /**
     * @return Chunk row index for the chunks intersecting the line given by
     * y-coordinate y.
     * @param y The y-coordinate of the line for which the chunk row index is
     *   returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkRowIndex(
            double y,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkRowIndex( y );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkRowIndex(
                        y,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Chunk row index for the chunks intersecting the line given by
     * y-coordinate y.
     * @param y The y-coordinate of the line for which the chunk row index is
     *   returned.
     */
    protected final int getChunkRowIndex(
            double y ) {
        return getChunkRowIndex( getCellRowIndex( y ) );
    }
    
    /**
     * @return Chunk row index for the chunk intersecting the cells with
     * cell row index _CellRowIndex.
     * @param _CellRowIndex The cell row index of the cells thats chunk row index
     *   is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkRowIndex(
            long _CellRowIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkRowIndex( _CellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkRowIndex(
                        _CellRowIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Chunk row index for the chunk intersecting the cells with
     * cell row index _CellRowIndex.
     * @param _CellRowIndex The cell row index of the cells thats chunk row index
     *   is returned.
     */
    protected final int getChunkRowIndex(
            long _CellRowIndex ) {
        return ( int ) ( _CellRowIndex / ( long ) this._ChunkNRows );
    }
    
    /**
     * @return Cell row Index for the cells that intersect the line with
     * y-coordinate y.
     * @param y The y-coordinate of the line thats cell row index is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellRowIndex(
            double y,
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = getCellRowIndex( y );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellRowIndex(
                        y,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell row Index for the cells that intersect the line with
     * y-coordinate y.
     * @param y The y-coordinate of the line thats cell row index is returned.
     */
    protected final long getCellRowIndex(
            double y ) {
        return getCellRowIndex( BigDecimal.valueOf( y ) );
    }
    
    /**
     * @return Cell row Index for the cells that intersect the line with
     * y-coordinate yBigDecimal.
     * @param yBigDecimal The y-coordinate of the line for which the cell row index
     *   is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellRowIndex(
            BigDecimal yBigDecimal,
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = getCellRowIndex( yBigDecimal );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellRowIndex(
                        yBigDecimal,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell row Index for the cells that intersect the line with
     * y-coordinate yBigDecimal.
     * @param yBigDecimal The y-coordinate of the line for which the cell row index
     *   is returned.
     */
    protected final long getCellRowIndex(
            BigDecimal yBigDecimal ) {
        int scale = Math.max(
                yBigDecimal.scale(),
                this._DimensionsScale );
        BigDecimal _BigDecimal0 = yBigDecimal.subtract(
                this._Dimensions[ 2 ] );
        BigDecimal _BigDecimal1 = _BigDecimal0.divide(
                this._Dimensions[ 0 ],
                scale,
                BigDecimal.ROUND_DOWN );
        return _BigDecimal1.longValue();
    }
    
    /**
     * @param _ChunkRowIndex
     * @param chunkCellRowIndex
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return long CellRowIndex, the cell row index for the cells in
     * chunk row index _ChunkRowIndex
     * chunk cell row index chunkCellRowIndex.
     */
    public final long getCellRowIndex(
            int _ChunkRowIndex,
            int chunkCellRowIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = getCellRowIndex(
                    _ChunkRowIndex,
                    chunkCellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellRowIndex(
                        _ChunkRowIndex,
                        chunkCellRowIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @param _ChunkRowIndex
     * @param chunkCellRowIndex
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param _ChunkID This is a ChunkID for those AbstractGrid2DSquareCells not
     * to be swapped if possible when an OutOfMemoryError is encountered.
     * @return long CellRowIndex, the cell row index for the cells in
     * chunk row index _ChunkRowIndex
     * chunk cell row index chunkCellRowIndex.
     */
    public final long getCellRowIndex(
            int _ChunkRowIndex,
            int chunkCellRowIndex,
            boolean _HandleOutOfMemoryError,
            ChunkID _ChunkID ) {
        try {
            long result = getCellRowIndex(
                    _ChunkRowIndex,
                    chunkCellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _SwapToFileGrid2DSquareCellChunkExcept( _ChunkID ) == null ) {
                    _SwapToFileGrid2DSquareCellChunk();
                }
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellRowIndex(
                        _ChunkRowIndex,
                        chunkCellRowIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * If OutOfMemoryError is caught then an attempt is made to swap an
     * AbstractGrid2DSquareCellChunk not in _Grid2DSquareCell_ChunkIDHashSet.
     * If this is not done then any AbstractGrid2DSquareCellChunk is swapped and
     * a warning is printed to sout
     * @param _ChunkRowIndex
     * @param chunkCellRowIndex
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return long CellRowIndex, the cell row index for the cells in
     * chunk row index _ChunkRowIndex
     * chunk cell row index chunkCellRowIndex.
     */
    public final long getCellRowIndex(
            int _ChunkRowIndex,
            int chunkCellRowIndex,
            boolean _HandleOutOfMemoryError,
            HashMap _Grid2DSquareCell_ChunkIDHashSet ) {
        try {
            long result = getCellRowIndex(
                    _ChunkRowIndex,
                    chunkCellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap _SwapToFileGrid2DSquareCellChunkHashMap = _SwapToFileGrid2DSquareCellChunkExcept(
                        _Grid2DSquareCell_ChunkIDHashSet );
                if ( _SwapToFileGrid2DSquareCellChunkHashMap.isEmpty() ) {
                    _SwapToFileGrid2DSquareCellChunk();
                    System.out.println( "Warning, unable to _SwapToFileGrid2DSquareCellChunkExcept(HashMap) in " + this.getClass() + ".getCellRowIndex(int,int,boolean,HashMap)." );
                }
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellRowIndex(
                        _ChunkRowIndex,
                        chunkCellRowIndex,
                        _HandleOutOfMemoryError,
                        _Grid2DSquareCell_ChunkIDHashSet );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Cell column index for the cells in chunk column index
     * _ChunkColIndex chunk cell column index chunkCellColIndex.
     * @param _ChunkRowIndex
     * @param chunkCellRowIndex
     */
    protected final long getCellRowIndex(
            int _ChunkRowIndex,
            int chunkCellRowIndex ) {
        return ( ( long ) _ChunkRowIndex * ( long ) this._ChunkNRows ) + ( long ) chunkCellRowIndex;
    }
    
    /**
     * @return Chunk cell row Index of the cells that intersects the line with
     * y-coordinate y.
     * @param y The y-coordinate of the line for which the chunk cell row index is
     *   returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellRowIndex(
            double y,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkCellRowIndex( y );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkCellRowIndex(
                        y,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Chunk cell row Index of the cells that intersects the line with
     * y-coordinate y.
     * @param y The y-coordinate of the line for which the chunk cell row index is
     *   returned.
     */
    protected final int getChunkCellRowIndex(
            double y ) {
        return getChunkCellRowIndex( getCellRowIndex( y ) );
    }
    
    /**
     * @return Chunk cell row index of the cells with cell row index equal to
     * _CellRowIndex.
     * @param _CellRowIndex The cell row index of the cells that chunk cell row
     *   index is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellRowIndex(
            long _CellRowIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            int result = getChunkCellRowIndex( _CellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getChunkCellRowIndex(
                        _CellRowIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Chunk cell row index of the cells with cell row index equal to
     * _CellRowIndex.
     * @param _CellRowIndex The cell row index of the cells that chunk cell row
     *   index is returned.
     */
    protected final int getChunkCellRowIndex(
            long _CellRowIndex ) {
        return ( int ) ( _CellRowIndex - ( ( _CellRowIndex / this._ChunkNRows ) * this._ChunkNRows ) );
    }
    
    /**
     * @return CellID of the cell given by cell row index _CellRowIndex, cell
     * column index _CellColIndex. A CellID is returned even if that cell would not
     * be in the grid.
     * @param _CellRowIndex The cell row index.
     * @param _CellColIndex The cell column index.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID getCellID(
            long _CellRowIndex,
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ){
        try {
            CellID result = getCellID(
                    _CellRowIndex,
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellID(
                        _CellRowIndex,
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return CellID of the cell given by cell row index _CellRowIndex, cell
     * column index _CellColIndex. A CellID is returned even if that cell would not
     * be in the grid.
     * @param _CellRowIndex The cell row index.
     * @param _CellColIndex The cell column index.
     */
    protected final CellID getCellID(
            long _CellRowIndex,
            long _CellColIndex ) {
        return new CellID( this._NCols, _CellRowIndex, _CellColIndex );
    }
    
    /**
     * @return CellID of the cell given by x-coordinate x, y-coordinate y even
     * if that cell would not be in the grid.
     * @param x The x-coordinate.
     * @param y The y-coordinate.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID getCellID(
            double x,
            double y,
            boolean _HandleOutOfMemoryError ) {
        try {
            CellID result = getCellID(
                    x,
                    y );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellID(
                        x,
                        y,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a CellID of the cell given by x-coordinate x, y-coordinate y even
     * if that cell would not be in the grid.
     * @param x The x-coordinate.
     * @param y The y-coordinate.
     */
    protected final CellID getCellID(
            double x,
            double y ) {
        return new CellID(
                this._NCols,
                getCellRowIndex( y ),
                getCellColIndex( x ) );
    }
    
    /**
     * Attempts to write this instance to Files located in the _Directory returned by
     * get_Directory().
     * @param swapToFileCache Iff true then this._Grid2DSquareCellChunks
     *   is written to new File( get_Directory(), "cache" ).
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFile(
            boolean swapToFileCache,
            boolean _HandleOutOfMemoryError )
            throws IOException {
        try {
            writeToFile( swapToFileCache );
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                writeToFile(
                        swapToFileCache,
                        _HandleOutOfMemoryError );
                init_MemoryReserve( _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write this instance to Files located in the _Directory returned by
     * get_Directory().
     * @param swapToFileCache Iff true then this._Grid2DSquareCellChunks
     *   is written to new File(get_Directory(),"cache").
     */
    protected void writeToFile(
            boolean swapToFileCache )
            throws IOException {
        try {
            writeToFileGrid2DSquareCellChunks();
            if ( swapToFileCache ) {
                // Write out thisCache
                File file = new File(
                        this._Directory,
                        "cache" );
                ObjectOutputStream oos = new ObjectOutputStream(
                        new BufferedOutputStream(
                        new FileOutputStream( file ) ) );
                oos.writeObject( this._Grid2DSquareCellChunks );
                oos.flush();
                oos.close();
            }
            // Write out this.
            File file = new File(
                    this._Directory,
                    "thisFile" );
            ObjectOutputStream oos = new ObjectOutputStream(
                    new BufferedOutputStream(
                    new FileOutputStream( file ) ) );
            oos.writeObject( this );
            oos.flush();
            oos.close();
        } catch ( IOException ioe0 ) {
            System.err.println( ioe0.getMessage() );
            throw ioe0;
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks. The first
     * AbstractGrid2DSquareCellChunk attempted to be written is that with a
     * chunk row index of 0, and a chunk column index of 0.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk which was written to file.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunk(
            boolean _HandleOutOfMemoryError ) {
        try {
            ChunkID result = writeToFileGrid2DSquareCellChunk();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return writeToFileGrid2DSquareCellChunk(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks. The first
     * AbstractGrid2DSquareCellChunk attempted to be written is that with a
     * chunk row index of 0, and a chunk column index of 0.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk which was swapped or null.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunk() {
        int _ChunkRowIndex;
        int _ChunkColIndex;
        ChunkID _ChunkID;
        for ( _ChunkRowIndex = 0; _ChunkRowIndex < this._NChunkRows; _ChunkRowIndex ++ ) {
            for ( _ChunkColIndex = 0; _ChunkColIndex < this._NChunkCols; _ChunkColIndex ++ ) {
                _ChunkID = new ChunkID(
                        this._NChunkCols,
                        _ChunkRowIndex,
                        _ChunkColIndex );
                if ( writeToFileGrid2DSquareCellChunk( _ChunkID ) ) {
                    return _ChunkID;
                }
            }
        }
        return null;
    }
    
    /**
     * Attempts to write to File a seriailized version of the
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks given by;
     * chunk row index _ChunkRowIndex, chunk column index _ChunkColIndex.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk to be written.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk to be written.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return True if AbstractGrid2DSquareCellChunk on file is up to date.
     */
    public final boolean writeToFileGrid2DSquareCellChunk(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = writeToFileGrid2DSquareCellChunk(
                    _ChunkRowIndex,
                    _ChunkColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return writeToFileGrid2DSquareCellChunk(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of the
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks given by;
     * chunk row index _ChunkRowIndex, chunk column index _ChunkColIndex.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk to be written.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk to be written.
     * @return True if AbstractGrid2DSquareCellChunk on file is up to date.
     */
    protected final boolean writeToFileGrid2DSquareCellChunk(
            int _ChunkRowIndex,
            int _ChunkColIndex ) {
        ChunkID _ChunkID = new ChunkID(
                this._NChunkCols,
                _ChunkRowIndex,
                _ChunkColIndex );
        return writeToFileGrid2DSquareCellChunk( _ChunkID );
    }
    
    /**
     * Attempts to write to File a seriailized version of the
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks given by _ChunkID.
     * @param _ChunkID The _ChunkID of the AbstractGrid2DSquareCellChunk to be written.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return True if AbstractGrid2DSquareCellChunk on file is up to date.
     */
    public final boolean writeToFileGrid2DSquareCellChunk(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = writeToFileGrid2DSquareCellChunk( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return writeToFileGrid2DSquareCellChunk(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of the
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks given by _ChunkID.
     * @param _ChunkID The _ChunkID of the AbstractGrid2DSquareCellChunk to be written.
     * @return True if AbstractGrid2DSquareCellChunk on file is up to date.
     */
    protected final boolean writeToFileGrid2DSquareCellChunk(
            ChunkID _ChunkID ) {
        boolean result = true;
        Object object = this._Grid2DSquareCellChunks.get( _ChunkID );
        if ( object != null ) {
            AbstractGrid2DSquareCellChunk grid2DSquareCellChunk =
                    ( AbstractGrid2DSquareCellChunk ) object;
            if ( ! grid2DSquareCellChunk.getIsSwapUpToDate() ) {
                File file = new File(
                        this._Directory,
                        _ChunkID.getChunkRowIndex() + "_" + _ChunkID.getChunkColIndex() );
                file.getParentFile().mkdirs();
                try {
                    file.createNewFile();
                    ObjectOutputStream oos = new ObjectOutputStream(
                            new BufferedOutputStream(
                            new FileOutputStream( file ) ) );
                    oos.writeObject( grid2DSquareCellChunk );
                    oos.flush();
                    oos.close();
                } catch ( IOException ioe0 ) {
                    ioe0.printStackTrace();
                }
                System.gc();
                grid2DSquareCellChunk.setIsSwapUpToDate( true );
            }
        } else {
            result = false;
        }
        return result;
    }
    
    /**
     * Attempts to write to File seriailized versions of all
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFileGrid2DSquareCellChunks(
            boolean _HandleOutOfMemoryError ) {
        try {
            writeToFileGrid2DSquareCellChunks();
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                writeToFileGrid2DSquareCellChunks(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File seriailized versions of all
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks.
     */
    protected final void writeToFileGrid2DSquareCellChunks() {
        Iterator iterator = this._Grid2DSquareCellChunks.keySet().iterator();
        ChunkID _ChunkID;
        while ( iterator.hasNext() ) {
            _ChunkID = ( ChunkID ) iterator.next();
            if ( isInCache( _ChunkID ) ) {
                writeToFileGrid2DSquareCellChunk( _ChunkID );
            }
        }
    }
    
    /**
     * Attempts to write to File seriailized versions of those
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks that
     * have ChunkID equal to those in _ChunkIDs.
     * @param _ChunkIDs A HashSet containing the ChunkID of the
     *   AbstractGrid2DSquareCellChunk to be written to file.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFileGrid2DSquareCellChunks(
            HashSet _ChunkIDs,
            boolean _HandleOutOfMemoryError ) {
        try {
            writeToFileGrid2DSquareCellChunks( _ChunkIDs );
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunkExcept(
                        this,
                        _ChunkIDs );
                init_MemoryReserve(
                        this,
                        _ChunkIDs,
                        _HandleOutOfMemoryError );
                writeToFileGrid2DSquareCellChunks(
                        _ChunkIDs,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File seriailized versions of those
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks that
     * have ChunkID equal to those in _ChunkIDs.
     * @param _ChunkIDs A HashSet containing the ChunkID of the
     *   AbstractGrid2DSquareCellChunk to be written to file.
     */
    protected final void writeToFileGrid2DSquareCellChunks(
            HashSet _ChunkIDs ) {
        Iterator iterator = _ChunkIDs.iterator();
        ChunkID _ChunkID;
        while( iterator.hasNext() ) {
            _ChunkID = ( ChunkID ) iterator.next();
            if ( isInCache( _ChunkID ) ) {
                writeToFileGrid2DSquareCellChunk( _ChunkID );
            }
        }
    }
    
    /**
     * Attempts to write to File seriailized versions of those
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with ChunkID equal to _ChunkID.
     * @param _ChunkID The ChunkID of the chunk not to be written.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFileGrid2DSquareCellChunksExcept(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            writeToFileGrid2DSquareCellChunksExcept( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunkExcept(
                        this,
                        _ChunkID );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                writeToFileGrid2DSquareCellChunksExcept(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File seriailized versions of those
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with ChunkID equal to _ChunkID.
     * @param _ChunkID The ChunkID of the chunk not to be written.
     */
    protected final void writeToFileGrid2DSquareCellChunksExcept(
            ChunkID _ChunkID ) {
        int _ChunkRowIndex;
        int _ChunkColIndex;
        ChunkID _ChunkIDToswapToFile;
        for ( _ChunkRowIndex = 0; _ChunkRowIndex < this._NChunkRows; _ChunkRowIndex ++ ) {
            for ( _ChunkColIndex = 0; _ChunkColIndex < this._NChunkCols; _ChunkColIndex ++ ) {
                _ChunkIDToswapToFile = new ChunkID(
                        this._NChunkCols,
                        _ChunkRowIndex,
                        _ChunkColIndex );
                if (  isInCache( _ChunkIDToswapToFile ) &&
                        ! _ChunkIDToswapToFile.equals( _ChunkID ) ) {
                    writeToFileGrid2DSquareCellChunk(
                            _ChunkRowIndex,
                            _ChunkColIndex );
                }
            }
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with ChunkID equal to _ChunkID. The first chunk that is attempted
     * to be written is that with a chunk row index of 0, and a chunk column index
     * of 0.
     * @param _ChunkID The ChunkID of the chunk not to be swapToFiled.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk that is written or null.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunkExcept(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            ChunkID result = writeToFileGrid2DSquareCellChunkExcept( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunkExcept(
                        this,
                        _ChunkID );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return writeToFileGrid2DSquareCellChunkExcept(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with ChunkID equal to _ChunkID. The first chunk that is attempted
     * to be written is that with a chunk row index of 0, and a chunk column index
     * of 0.
     * @param _ChunkID The ChunkID of the chunk not to be swapToFiled.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk that is written or null.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunkExcept(
            ChunkID _ChunkID ) {
        return writeToFileGrid2DSquareCellChunkExcept(
                _ChunkID.getChunkRowIndex(),
                _ChunkID.getChunkColIndex() );
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with; chunk row index _ChunkRowIndex, chunk column index _ChunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk
     * row index of 0, and a chunk column index of 0.
     * @param _ChunkRowIndex The chunk row index.
     * @param _ChunkColIndex The chunk column index.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk that is written or null.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunkExcept(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            return writeToFileGrid2DSquareCellChunkExcept(
                    _ChunkRowIndex,
                    _ChunkColIndex );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            _ChunkRowIndex,
                            _ChunkColIndex );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return writeToFileGrid2DSquareCellChunkExcept(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with; chunk row index _ChunkRowIndex, chunk column index _ChunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk
     * row index of 0, and a chunk column index of 0.
     * @param _ChunkRowIndex The chunk row index.
     * @param _ChunkColIndex The chunk column index.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk that is written or null.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunkExcept(
            int _ChunkRowIndex,
            int _ChunkColIndex ) {
        int cri;
        int cci;
        ChunkID _ChunkID = null;
        for ( cri = 0; cri < this._NChunkRows; cri ++ ) {
            for ( cci = 0; cci< this._NChunkCols; cci ++ ) {
                if ( ! ( ( cri == _ChunkRowIndex ) && ( cci == _ChunkColIndex ) ) )  {
                    _ChunkID = new ChunkID(
                            this._NChunkCols,
                            cri,
                            cci );
                    if ( isInCache( _ChunkID ) ) {
                        writeToFileGrid2DSquareCellChunk(
                                cri,
                                cci );
                        return _ChunkID;
                    }
                }
            }
        }
        return _ChunkID;
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with; chunk row index _ChunkRowIndex, chunk column index _ChunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk row
     * index of _ChunkNRows - 1, and a chunk column index of _ChunkNCols - 1.
     * This is primarily used in constructors that read a data format that
     * @param _ChunkRowIndex The chunk row index.
     * @param _ChunkColIndex The chunk column index.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk that is written or null.
     * presents values in reverse row order.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            //return writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            //        _ChunkRowIndex,
            //        _ChunkColIndex );
            ChunkID result = writeToFileGrid2DSquareCellChunkExceptReverseOrder(
                    _ChunkRowIndex,
                    _ChunkColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    ChunkID _ChunkID = new ChunkID(
                            this._NChunkCols,
                            _ChunkRowIndex,
                            _ChunkColIndex );
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return writeToFileGrid2DSquareCellChunkExceptReverseOrder(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with; chunk row index _ChunkRowIndex, chunk column index _ChunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk row
     * index of _ChunkNRows - 1, and a chunk column index of _ChunkNCols - 1.
     * This is primarily used in constructors that read a data format that
     * presents values in reverse row order.
     * @param _ChunkRowIndex The chunk row index.
     * @param _ChunkColIndex The chunk column index.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk that is written or null.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            int _ChunkRowIndex,
            int _ChunkColIndex ) {
        ChunkID _ChunkID = null;
        int cri;
        int cci;
        for ( cri = this._NChunkRows - 1; cri > -1; cri -- ) {
            for ( cci = this._NChunkCols - 1; cci > -1; cci -- ) {
                if ( ! ( ( cri == _ChunkRowIndex ) && ( cci == _ChunkColIndex ) ) )  {
                    _ChunkID = new ChunkID(
                            this._NChunkCols,
                            cri,
                            cci );
                    if ( isInCache( _ChunkID ) ) {
                        writeToFileGrid2DSquareCellChunk(
                                cri,
                                cci );
                        return _ChunkID;
                    }
                }
            }
        }
        return _ChunkID;
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with; chunk row index _ChunkRowIndex, chunk column index _ChunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk row
     * index of _ChunkNRows - 1, and a chunk column index of _ChunkNCols - 1. This
     * is primarily used in constructors that read a data format that presents
     * values in reverse row order.
     * @param _ChunkID The ChunkID of the chunk not to be swapToFiled.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return ChunkID of the AbstractGrid2DSquareCellChunk that is written or null.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            //return writeToFileGrid2DSquareCellChunkExceptReverseOrder( _ChunkID );
            ChunkID result = writeToFileGrid2DSquareCellChunkExceptReverseOrder( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _HandleOutOfMemoryError ) {
                    _SwapToFileGrid2DSquareCellChunkExcept(
                            this,
                            _ChunkID );
                    init_MemoryReserve(
                            this,
                            _ChunkID,
                            _HandleOutOfMemoryError );
                }
                return writeToFileGrid2DSquareCellChunkExceptReverseOrder(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File a seriailized version of a
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * that with; chunk row index _ChunkRowIndex, chunk column index _ChunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk row
     * index of _ChunkNRows - 1, and a chunk column index of _ChunkNCols - 1. This
     * is primarily used in constructors that read a data format that presents
     * values in reverse row order.
     * @param _ChunkID The ChunkID of the chunk not to be swapToFiled.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            ChunkID _ChunkID ) {
        return writeToFileGrid2DSquareCellChunkExceptReverseOrder(
                _ChunkID.getChunkRowIndex(),
                _ChunkID.getChunkColIndex() );
    }
    
    /**
     * Attempts to write to File seriailized versions of those
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * those that have ChunkID in _ChunkIDs.
     * @param _ChunkIDs A hashSet containing ChunkIDs not to be swapToFiled.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFileGrid2DSquareCellChunksExcept(
            HashSet _ChunkIDs,
            boolean _HandleOutOfMemoryError ) {
        try {
            writeToFileGrid2DSquareCellChunksExcept( _ChunkIDs );
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                swapToFileGrid2DSquareCellChunkExcept( _ChunkIDs );
                writeToFileGrid2DSquareCellChunksExcept(
                        _ChunkIDs,
                        _HandleOutOfMemoryError );
                init_MemoryReserve(
                        this,
                        _ChunkIDs,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to File seriailized versions of those
     * AbstractGrid2DSquareCellChunk in this._Grid2DSquareCellChunks except
     * those that have ChunkID in _ChunkIDs.
     * @param _ChunkIDs A hashSet containing ChunkIDs not to be swapToFiled.
     */
    protected final void writeToFileGrid2DSquareCellChunksExcept(
            HashSet _ChunkIDs ) {
        int _ChunkRowIndex;
        int _ChunkColIndex;
        ChunkID _ChunkIDToswapToFile;
        for ( _ChunkRowIndex = 0; _ChunkRowIndex < this._NChunkRows; _ChunkRowIndex ++ ) {
            for ( _ChunkColIndex = 0; _ChunkColIndex < this._NChunkCols; _ChunkColIndex ++ ) {
                _ChunkIDToswapToFile = new ChunkID(
                        this._NChunkCols,
                        _ChunkRowIndex,
                        _ChunkColIndex );
                if ( ! ( isInCache( _ChunkIDToswapToFile ) &&
                        _ChunkIDs.contains( _ChunkIDToswapToFile ) ) ) {
                    writeToFileGrid2DSquareCellChunk(
                            _ChunkRowIndex,
                            _ChunkColIndex );
                }
            }
        }
    }
    
//    /**
//     * Attempts to write to file and clear from the cache a
//     * AbstractGrid2DSquareCellChunk in this.
//     * @return A HashMap with this as key and the
//     * AbstractGrid2DSquareCellChunk ChunkID swapped or null
//     * as value.
//     * @param _HandleOutOfMemoryError
//     *   If true then OutOfMemoryErrors are caught, swap operations are
//     *     initiated, then the method is re-called.
//     *   If false then OutOfMemoryErrors are caught and thrown.
//     */
//    public HashMap swapToFileGrid2DSquareCellChunk(
//            boolean _HandleOutOfMemoryError ) {
//        try {
//            HashMap result = swapToFileGrid2DSquareCellChunk();
//            _EnsureThereIsEnoughMemoryToContinue();
//            return result;
//        } catch ( OutOfMemoryError _OutOfMemoryError ) {
//            if ( _HandleOutOfMemoryError ) {
//                clear_MemoryReserve();
//                HashMap result = _SwapToFileGrid2DSquareCellChunk(
//                        _HandleOutOfMemoryError );
//                init_MemoryReserve( _HandleOutOfMemoryError );
//                return result;
//            } else {
//                throw _OutOfMemoryError;
//            }
//        }
//    }
    
    /**
     * Attempts to write to file and clear from the cache a
     * AbstractGrid2DSquareCellChunk in this.
     * This is the lowest level memory handling operation of this class.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return ChunkID swapped or null.
     */
    public ChunkID swapToFileGrid2DSquareCellChunk(
            boolean _HandleOutOfMemoryError ) {
        try {
            ChunkID result = swapToFileGrid2DSquareCellChunk();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
//                _SwapToFileGrid2DSquareCellChunkExcept( this );
//                init_MemoryReserve(
//                        this,
//                        _HandleOutOfMemoryError );
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve(
                        _HandleOutOfMemoryError );
                return swapToFileGrid2DSquareCellChunk(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
//    /**
//     * Attempts to write to file and clear from the cache a
//     * AbstractGrid2DSquareCellChunk in this.
//     * @return A HashMap with this as key and the
//     * AbstractGrid2DSquareCellChunk ChunkID swapped or null
//     * as value.
//     */
//    protected final HashMap swapToFileGrid2DSquareCellChunk() {
//        HashMap result = new HashMap();
//        ChunkID _ChunkID = writeToFileGrid2DSquareCellChunk();
//        clearFromCacheGrid2DSquareCellChunk( _ChunkID );
//        result.put( this, _ChunkID );
//        return result;
//    }
    
    /**
     * Attempts to write to file and clear from the cache a
     * AbstractGrid2DSquareCellChunk in this.
     *
     *
     * @return ChunkID swapped or null.
     */
    protected final ChunkID swapToFileGrid2DSquareCellChunk() {
        ChunkID _ChunkID = writeToFileGrid2DSquareCellChunk();
        clearFromCacheGrid2DSquareCellChunk( _ChunkID );
        return _ChunkID;
    }
    
    /**
     * Attempts to write to file and clear from the cache a AbstractGrid2DSquareCellChunk
     * in this._Grid2DSquareCells.
     *
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The ChunkID of AbstractGrid2DSquareCellChunk swapped or null.
     */
    public final ChunkID swapToFileGrid2DSquareCellChunkExcept(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            ChunkID result = swapToFileGrid2DSquareCellChunkExcept( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunkExcept(
                        this,
                        _ChunkID );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return swapToFileGrid2DSquareCellChunkExcept(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache a AbstractGrid2DSquareCellChunk
     * in this._Grid2DSquareCells.
     * @return The ChunkID of AbstractGrid2DSquareCellChunk swapped or null.
     */
    protected final ChunkID swapToFileGrid2DSquareCellChunkExcept(
            ChunkID _ChunkID ) {
        int cri;
        int cci;
        ChunkID thisChunkID = null;
        for ( cri = 0; cri < this._NChunkRows; cri ++ ) {
            for ( cci = 0; cci< this._NChunkCols; cci ++ ) {
                thisChunkID = new ChunkID(
                        this._NChunkCols,
                        cri,
                        cci );
                if ( ! thisChunkID.equals( _ChunkID ) ) {
                    if ( isInCache( _ChunkID ) ) {
                        writeToFileGrid2DSquareCellChunk( _ChunkID );
                        clearFromCacheGrid2DSquareCellChunk( _ChunkID );
                        return thisChunkID;
                    }
                }
            }
        }
        return thisChunkID;
    }
    
    /**
     * Attempts to write to file and clear from the cache all
     * Grid2DSquareCellChunkAbstracts in this except that with
     * ChunkID _ChunkID.
     *
     *
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of AbstractGrid2DSquareCellChunk swapped.
     */
    public final HashSet swapToFileGrid2DSquareCellChunksExcept(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashSet result = swapToFileGrid2DSquareCellChunksExcept( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunkExcept(
                        this,
                        _ChunkID );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return swapToFileGrid2DSquareCellChunksExcept(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache all
     * Grid2DSquareCellChunkAbstracts in this except that with
     * ChunkID _ChunkID.
     * @param _ChunkID ChunkID of the AbstractGrid2DSquareCellChunk not to be swapped.
     * @return A HashSet with the ChunkIDs of those
     * AbstractGrid2DSquareCellChunk swapped.
     */
    protected final HashSet swapToFileGrid2DSquareCellChunksExcept(
            ChunkID _ChunkID ) {
        HashSet result = new HashSet();
        int cri;
        int cci;
        ChunkID _ChunkIDToSwap = null;
        for ( cri = 0; cri < this._NChunkRows; cri ++ ) {
            for ( cci = 0; cci< this._NChunkCols; cci ++ ) {
                _ChunkIDToSwap = new ChunkID(
                        this._NChunkCols,
                        cri,
                        cci );
                if ( ! _ChunkID.equals( _ChunkIDToSwap ) ) {
                    if ( isInCache( _ChunkIDToSwap ) ) {
                        writeToFileGrid2DSquareCellChunk( _ChunkIDToSwap );
                        clearFromCacheGrid2DSquareCellChunk( _ChunkIDToSwap );
                        result.add( _ChunkIDToSwap );
                    }
                }
            }
        }
        return result;
    }
    
    /**
     * Attempts to write to file and clear from the cache all
     * Grid2DSquareCellChunkAbstracts in this except that with
     * ChunkID _ChunkIDs.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of AbstractGrid2DSquareCellChunk swapped.
     */
    public final HashSet swapToFileGrid2DSquareCellChunksExcept(
            HashSet _ChunkIDs,
            boolean _HandleOutOfMemoryError ) {
        try {
            HashSet result = swapToFileGrid2DSquareCellChunksExcept( _ChunkIDs );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                HashMap _HashMap = _SwapToFileGrid2DSquareCellChunkExcept(
                        this,
                        _ChunkIDs );
                if ( _HashMap.containsKey( this ) ) {
                    HashSet result = ( HashSet ) _HashMap.get( this );
                    result.addAll(
                            swapToFileGrid2DSquareCellChunksExcept(
                            _ChunkIDs,
                            _HandleOutOfMemoryErrorFalse ) );
                    init_MemoryReserve(
                            this,
                            _ChunkIDs,
                            _HandleOutOfMemoryError );
                    return result;
                } else {
                    init_MemoryReserve(
                            this,
                            _ChunkIDs,
                            _HandleOutOfMemoryError );
                    return swapToFileGrid2DSquareCellChunksExcept(
                            _ChunkIDs,
                            _HandleOutOfMemoryErrorFalse );
                }
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache all
     * Grid2DSquareCellChunkAbstracts in this except that with
     * ChunkID _ChunkID.
     * @param _ChunkIDs HashSet of AbstractGrid2DSquareCellChunk.ChunkIDs
     * not to be swapped.
     * @return A HashSet with the ChunkIDs of those
     * AbstractGrid2DSquareCellChunk swapped.
     */
    protected final HashSet swapToFileGrid2DSquareCellChunksExcept(
            HashSet _ChunkIDs ) {
        HashSet result = new HashSet();
        int cri;
        int cci;
        ChunkID _ChunkIDToSwap = null;
        for ( cri = 0; cri < this._NChunkRows; cri ++ ) {
            for ( cci = 0; cci< this._NChunkCols; cci ++ ) {
                _ChunkIDToSwap = new ChunkID(
                        this._NChunkCols,
                        cri,
                        cci );
                if ( ! _ChunkIDs.contains( _ChunkIDToSwap ) ) {
                    if ( isInCache( _ChunkIDToSwap ) ) {
                        writeToFileGrid2DSquareCellChunk( _ChunkIDToSwap );
                        clearFromCacheGrid2DSquareCellChunk( _ChunkIDToSwap );
                        result.add( _ChunkIDToSwap );
                    }
                }
            }
        }
        return result;
    }
    
    /**
     * Attempts to write to file and clear from the cache a AbstractGrid2DSquareCellChunk
     * in this._Grid2DSquareCells.
     *
     *
     * @param _ChunkIDs
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The ChunkID of the AbstractGrid2DSquareCellChunk swapped or null.
     */
    public final ChunkID swapToFileGrid2DSquareCellChunkExcept(
            HashSet _ChunkIDs,
            boolean _HandleOutOfMemoryError ) {
        try {
            ChunkID result = swapToFileGrid2DSquareCellChunkExcept( _ChunkIDs );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunkExcept(
                        this,
                        _ChunkIDs );
                init_MemoryReserve(
                        this,
                        _ChunkIDs,
                        _HandleOutOfMemoryError );
                return swapToFileGrid2DSquareCellChunkExcept(
                        _ChunkIDs,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache a AbstractGrid2DSquareCellChunk
     * in this._Grid2DSquareCells.
     *
     *
     *
     * @return The ChunkID of the AbstractGrid2DSquareCellChunk swapped or null.
     */
    protected final ChunkID swapToFileGrid2DSquareCellChunkExcept(
            HashSet _ChunkIDs ) {
        int cri;
        int cci;
        ChunkID _ChunkID = null;
        for ( cri = 0; cri < this._NChunkRows; cri ++ ) {
            for ( cci = 0; cci< this._NChunkCols; cci ++ ) {
                _ChunkID = new ChunkID(
                        this._NChunkCols,
                        cri,
                        cci );
                if ( ! _ChunkIDs.contains( _ChunkID ) ) {
                    if ( isInCache( _ChunkID ) ) {
                        writeToFileGrid2DSquareCellChunk( _ChunkID );
                        clearFromCacheGrid2DSquareCellChunk( _ChunkID );
                        return _ChunkID;
                    }
                }
            }
        }
        return _ChunkID;
    }
    
    /**
     * Attempts to swap seriailsed version of all Grid2DSquareCellChunks in this only.
     * To swap seriailsed version of all Grid2DSquareCellChunks in grid2DsquareCells use
     * <code>swapToFileGrid2DSquareCellGrid2DSquareCellChunks(boolean)</code>
     * Swapping involves writing to files and then clearing them from the cache.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of AbstractGrid2DSquareCellChunk swapped.
     */
    public final HashSet swapToFileGrid2DSquareCellChunks(
            boolean _HandleOutOfMemoryError ) {
        try {
            HashSet result = swapToFileGrid2DSquareCellChunks();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return swapToFileGrid2DSquareCellChunks(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to swap seriailsed version of all _Grid2DSquareCellChunks. This involves
     * writing them to files and then clearing them from the cache.
     * @return The number of AbstractGrid2DSquareCellChunk swapped.
     */
    protected HashSet swapToFileGrid2DSquareCellChunks() {
//        //Using Iterator as follows risks concurrent modification exception!
//        Iterator iterator = this._Grid2DSquareCellChunks.entrySet().iterator();
//        ChunkID _ChunkID;
//        while ( iterator.hasNext() ) {
//            _ChunkID = ( ChunkID ) ( ( Map.Entry ) iterator.next() ).getKey();
//            writeToFileGrid2DSquareCellChunk( _ChunkID );
//            iterator.remove();
//            System.gc();
//        }
        HashSet result = new HashSet();
        int cri;
        int cci;
        ChunkID _ChunkID;
        for ( cri = 0; cri < this._NChunkRows; cri ++ ) {
            for ( cci = 0; cci< this._NChunkCols; cci ++ ) {
                _ChunkID = new ChunkID(
                        this._NChunkCols,
                        cri,
                        cci );
                if ( isInCache( _ChunkID ) ) {
                    writeToFileGrid2DSquareCellChunk( _ChunkID );
                    clearFromCacheGrid2DSquareCellChunk( _ChunkID );
                    result.add( _ChunkID );
                }
            }
        }
        if ( result.isEmpty() ) {
            return null;
        }
        return result;
    }
    
    /**
     * Attempts to swap seriailsed version of all AbstractGrid2DSquareCellChunk from
     * (cri0, cci0) to (cri1, cci1) in row major order. This involves writing
     * them to files and then clearing them from the cache.
     * @param cri0 The chunk row index of the first AbstractGrid2DSquareCellChunk to be
     *   swapped.
     * @param cci0 The chunk column index of the first AbstractGrid2DSquareCellChunk to
     *   be swapped.
     * @param cri1 The chunk row index of the last AbstractGrid2DSquareCellChunk to be
     *   swapped.
     * @param cci1 The chunk column index of the last AbstractGrid2DSquareCellChunk to be
     *   swapped.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of AbstractGrid2DSquareCellChunk swapped.
     */
    public final long swapToFileGrid2DSquareCellChunks(
            int cri0,
            int cci0,
            int cri1,
            int cci1,
            boolean _HandleOutOfMemoryError ) {
        try {
            long result = swapToFileGrid2DSquareCellChunks(
                    cri0,
                    cci0,
                    cri1,
                    cci1 );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return swapToFileGrid2DSquareCellChunks(
                        cri0,
                        cci0,
                        cri1,
                        cci1,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to swap seriailsed version of all AbstractGrid2DSquareCellChunk from
     * (cri0, cci0) to (cri1, cci1) in row major order. This involves writing
     * them to files and then clearing them from the cache.
     * @param cri0 The chunk row index of the first Grid2DSquareCellChunks to be
     *   swapped.
     * @param cci0 The chunk column index of the first Grid2DSquareCellChunks to
     *   be swapped.
     * @param cri1 The chunk row index of the last Grid2DSquareCellChunks to be
     *   swapped.
     * @param cci1 The chunk column index of the last Grid2DSquareCellChunks to
     *   be swapped.
     * @return The number of AbstractGrid2DSquareCellChunk swapped.
     */
    protected final long swapToFileGrid2DSquareCellChunks(
            int cri0,
            int cci0,
            int cri1,
            int cci1 ) {
        ChunkID _ChunkID = null;
        long nGrid2DSquareCellChunksSwapped = 0L;
        if ( cri0 != cri1 ) {
            for ( int cci = cci0; cci < this._NChunkCols; cci ++ ) {
                _ChunkID = new ChunkID(
                        this._NChunkCols,
                        cri0,
                        cci );
                if ( isInCache( _ChunkID ) ) {
                    writeToFileGrid2DSquareCellChunk( _ChunkID );
                    clearFromCacheGrid2DSquareCellChunk( _ChunkID );
                    nGrid2DSquareCellChunksSwapped ++;
                }
            }
            for ( int cri = cri0 + 1; cri < cri1; cri ++ ) {
                for ( int cci = 0; cci < this._NChunkCols; cci ++ ) {
                    _ChunkID = new ChunkID(
                            this._NChunkCols,
                            cri,
                            cci );
                    if ( isInCache( _ChunkID ) ) {
                        writeToFileGrid2DSquareCellChunk( _ChunkID );
                        clearFromCacheGrid2DSquareCellChunk( _ChunkID );
                        nGrid2DSquareCellChunksSwapped ++;
                    }
                }
            }
            for ( int cci = 0; cci < cci1; cci ++ ) {
                _ChunkID = new ChunkID(
                        this._NChunkCols,
                        cri1,
                        cci );
                if ( isInCache( _ChunkID ) ) {
                    writeToFileGrid2DSquareCellChunk( _ChunkID );
                    clearFromCacheGrid2DSquareCellChunk( _ChunkID );
                    nGrid2DSquareCellChunksSwapped ++;
                }
            }
        } else {
            for ( int cci = cci0; cci < cci1 + 1; cci ++ ) {
                _ChunkID = new ChunkID(
                        this._NChunkCols,
                        cri0,
                        cci );
                if ( isInCache( _ChunkID ) ) {
                    writeToFileGrid2DSquareCellChunk( _ChunkID );
                    clearFromCacheGrid2DSquareCellChunk( _ChunkID );
                    nGrid2DSquareCellChunksSwapped ++;
                }
            }
        }
        return nGrid2DSquareCellChunksSwapped;
    }
    
    /**
     * @return true iff grid2DSquareCellChunk given by _ChunkID is in memory.
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk tested to see if
     *   it is swapToFiled.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInCache(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = isInCache( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                swapToFileGrid2DSquareCellChunkExcept( _ChunkID );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
                return isInCache(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return true iff grid2DSquareCellChunk given by _ChunkID is swapToFiled. This
     * must be an upToDate swapToFile.
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk tested to see if
     *   it is swapToFiled.
     */
    protected final boolean isInCache(
            ChunkID _ChunkID ) {
        return this._Grid2DSquareCellChunks.containsKey( _ChunkID );
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory. This is usually only
     * done after the equivallent of swapToFileChunk(ChunkID) has been called.
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk to be cleared.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunk(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunk( _ChunkID );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                clearFromCacheGrid2DSquareCellChunk( _ChunkID );
                init_MemoryReserve( _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory. This is usually only
     * done after the equivallent of swapToFileChunk(ChunkID) has been called.
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunk(
            ChunkID _ChunkID ) {
        this._Grid2DSquareCellChunks.remove( _ChunkID );
        System.gc();
    }
    
    /**
     * For releasing all AbstractGrid2DSquareCellChunk in
     * this._Grid2DSquareCellChunks.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunks(
            boolean _HandleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunks();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                clearFromCacheGrid2DSquareCellChunks();
                init_MemoryReserve( _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * For releasing all AbstractGrid2DSquareCellChunk in
     * this._Grid2DSquareCellChunks.
     */
    protected final void clearFromCacheGrid2DSquareCellChunks() {
        Iterator iterator = this._Grid2DSquareCellChunks.entrySet().iterator();
        while ( iterator.hasNext() ) {
            iterator.next();
            iterator.remove();
            System.gc();
        }
    }
    
    /**
     * Clears Grid2DSquareCellChunkAbstracts with ChunkIDs in _ChunkIDs.
     * @param _ChunkIDs A HashSet containing the ChunkIDs of the
     *   Grid2DSquareCellChunkAbstracts to be cleared.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunks(
            HashSet _ChunkIDs,
            boolean _HandleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunks( _ChunkIDs );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ){
                clear_MemoryReserve();
                clearFromCacheGrid2DSquareCellChunks( _ChunkIDs );
                init_MemoryReserve( _HandleOutOfMemoryError );
            }
        }
    }
    
    /**
     * Clears Grid2DSquareCellChunkAbstracts with ChunkIDs in _ChunkIDs.
     * @param _ChunkIDs A HashSet containing the ChunkIDs of the
     *   Grid2DSquareCellChunkAbstracts to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunks(
            HashSet _ChunkIDs ) {
        Iterator iterator = _ChunkIDs.iterator();
        ChunkID _ChunkIDToClear;
        while( iterator.hasNext() ) {
            _ChunkIDToClear = ( ChunkID ) iterator.next();
            clearFromCacheGrid2DSquareCellChunk( _ChunkIDToClear );
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by _ChunkID.
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunkExcept(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunkExcept( _ChunkID );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                clearFromCacheGrid2DSquareCellChunkExcept( _ChunkID );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by _ChunkID.
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunkExcept(
            ChunkID _ChunkID ) {
        int _ChunkRowIndex;
        int _ChunkColIndex;
        ChunkID _ChunkIDToClear;
        for ( _ChunkRowIndex = 0; _ChunkRowIndex < this._NChunkRows; _ChunkRowIndex ++ ) {
            for ( _ChunkColIndex = 0; _ChunkColIndex < this._NChunkCols; _ChunkColIndex ++ ) {
                _ChunkIDToClear = new ChunkID(
                        this._NChunkCols,
                        _ChunkRowIndex,
                        _ChunkColIndex );
                if ( ! _ChunkIDToClear.equals( _ChunkID ) ) {
                    clearFromCacheGrid2DSquareCellChunk( _ChunkIDToClear );
                    return;
                }
            }
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by _ChunkID.
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunksExcept(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunksExcept( _ChunkID );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                clearFromCacheGrid2DSquareCellChunksExcept( _ChunkID );
                init_MemoryReserve(
                        this,
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by _ChunkID.
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunksExcept(
            ChunkID _ChunkID ) {
        int _ChunkRowIndex;
        int _ChunkColIndex;
        ChunkID _ChunkIDToswapToFile;
        for ( _ChunkRowIndex = 0; _ChunkRowIndex < this._NChunkRows; _ChunkRowIndex ++ ) {
            for ( _ChunkColIndex = 0; _ChunkColIndex < this._NChunkCols; _ChunkColIndex ++ ) {
                _ChunkIDToswapToFile = new ChunkID(
                        _NChunkCols,
                        _ChunkRowIndex,
                        _ChunkColIndex );
                if ( ! _ChunkIDToswapToFile.equals( _ChunkID ) ) {
                    clearFromCacheGrid2DSquareCellChunk( _ChunkIDToswapToFile );
                }
            }
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except those with
     * ChunkIDs in _ChunkIDs.
     *
     * @param _ChunkIDs The HashSet of ChunkIDs of the _Grid2DSquareCellChunks not
     *   to be cleared.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunksExcept(
            HashSet _ChunkIDs,
            boolean _HandleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunksExcept( _ChunkIDs );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                clearFromCacheGrid2DSquareCellChunksExcept( _ChunkIDs );
                init_MemoryReserve(
                        this,
                        _ChunkIDs,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except those with
     * ChunkIDs in _ChunkIDs.
     *
     * @param _ChunkIDs The HashSet of ChunkIDs of the _Grid2DSquareCellChunks not
     *   to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunksExcept(
            HashSet _ChunkIDs ) {
        int _ChunkRowIndex;
        int _ChunkColIndex;
        ChunkID _ChunkIDToClear;
        for ( _ChunkRowIndex = 0; _ChunkRowIndex < this._NChunkRows; _ChunkRowIndex ++ ) {
            for ( _ChunkColIndex = 0; _ChunkColIndex < this._NChunkCols; _ChunkColIndex ++ ) {
                _ChunkIDToClear = new ChunkID( _NChunkCols, _ChunkRowIndex, _ChunkColIndex );
                if ( ! _ChunkIDs.contains( _ChunkIDToClear ) ) {
                    clearFromCacheGrid2DSquareCellChunk( _ChunkIDToClear );
                }
            }
        }
    }
    
    /**
     * Restores a grid2DSquareCellChunk from a swapped cache file.
     * @param _ChunkRowIndex The chunk row index of the grid2DSquareCellChunk
     *   to be restored.
     * @param _ChunkColIndex The chunk column index of the grid2DSquareCellChunk
     *   to be restored.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void loadIntoCacheChunk(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            loadIntoCacheChunk(
                    _ChunkRowIndex,
                    _ChunkColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                loadIntoCacheChunk(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Restores a grid2DSquareCellChunk from a swapped cache file.
     *
     *
     * @param _ChunkRowIndex The chunk row index of the grid2DSquareCellChunk
     *   to be restored.
     * @param _ChunkColIndex The chunk column index of the grid2DSquareCellChunk
     *   to be restored.
     */
    protected final void loadIntoCacheChunk(
            int _ChunkRowIndex,
            int _ChunkColIndex ) {
        boolean _HandleOutOfMemoryError = true;
        ChunkID _ChunkID = new ChunkID(
                this._NChunkCols,
                _ChunkRowIndex,
                _ChunkColIndex );
        loadIntoCacheChunk(
                _ChunkID,
                _HandleOutOfMemoryError );
    }
    
    /**
     * Attempts to load into the memory cache a AbstractGrid2DSquareCellChunk with ChunkID
     * @param _ChunkID The ChunkID of the grid2DSquareCellChunk to be restored.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void loadIntoCacheChunk(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            loadIntoCacheChunk( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                loadIntoCacheChunk(
                        _ChunkID,
                        _HandleOutOfMemoryErrorFalse );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Attempts to load into the memory cache AbstractGrid2DSquareCellChunk with
     * ChunkID equal to _ChunkID.
     *
     *
     * @param _ChunkID The ChunkID of the AbstractGrid2DSquareCellChunk to be restored.
     */
    protected final void loadIntoCacheChunk(
            ChunkID _ChunkID ) {
//        boolean _HandleOutOfMemoryError = true;
//        boolean isInCache = isInCache(
//                _ChunkID,
//                _HandleOutOfMemoryError );
        boolean isInCache = isInCache(
                _ChunkID );
        if ( ! isInCache ) {
            File file = new File(
                    this._Directory,
                    "" + _ChunkID.getChunkRowIndex() +
                    "_" + _ChunkID.getChunkColIndex() );
            Object object0 = loadObject(
                    file );
//            Object object0 = loadObject(
//                    file,
//                    _HandleOutOfMemoryError );
            if ( this.getClass() == Grid2DSquareCellInt.class ) {
                if ( object0.getClass() == Grid2DSquareCellIntChunk64CellMap.class ) {
                    Grid2DSquareCellIntChunk64CellMap grid2DSquareCellIntChunk64CellMap =
                            ( Grid2DSquareCellIntChunk64CellMap ) object0;
                    grid2DSquareCellIntChunk64CellMap.initGrid2DSquareCell( this );
                    grid2DSquareCellIntChunk64CellMap.initChunkID( _ChunkID );
                    this._Grid2DSquareCellChunks.put(
                            _ChunkID,
                            grid2DSquareCellIntChunk64CellMap );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                    Grid2DSquareCellIntChunkArray grid2DSquareCellIntChunkArray =
                            ( Grid2DSquareCellIntChunkArray ) object0;
                    grid2DSquareCellIntChunkArray.initGrid2DSquareCell( this );
                    grid2DSquareCellIntChunkArray.initChunkID( _ChunkID );
                    this._Grid2DSquareCellChunks.put(
                            _ChunkID,
                            grid2DSquareCellIntChunkArray );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                    Grid2DSquareCellIntChunkMap grid2DSquareCellIntChunkMap =
                            ( Grid2DSquareCellIntChunkMap ) object0;
                    grid2DSquareCellIntChunkMap.initGrid2DSquareCell( this );
                    grid2DSquareCellIntChunkMap.initChunkID( _ChunkID );
                    this._Grid2DSquareCellChunks.put(
                            _ChunkID,
                            grid2DSquareCellIntChunkMap );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                    Grid2DSquareCellIntChunkJAI grid2DSquareCellIntChunkJAI =
                            ( Grid2DSquareCellIntChunkJAI ) object0;
                    grid2DSquareCellIntChunkJAI.initGrid2DSquareCell( this );
                    grid2DSquareCellIntChunkJAI.initChunkID( _ChunkID );
                    this._Grid2DSquareCellChunks.put(
                            _ChunkID,
                            grid2DSquareCellIntChunkJAI );
                    return;
                }
                System.err.println(
                        "Unrecognised type of Grid2DSquareCellIntChunkAbstract or null " +
                        this.getClass().getName() +
                        ".loadIntoCacheChunk( ChunkID( " + _ChunkID.toString() + " ) )" );
            } else {
                if ( object0.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                    Grid2DSquareCellDoubleChunk64CellMap grid2DSquareCellDoubleChunk64CellMap =
                            ( Grid2DSquareCellDoubleChunk64CellMap ) object0;
                    grid2DSquareCellDoubleChunk64CellMap.initGrid2DSquareCell( this );
                    grid2DSquareCellDoubleChunk64CellMap.initChunkID( _ChunkID );
                    this._Grid2DSquareCellChunks.put(
                            _ChunkID,
                            grid2DSquareCellDoubleChunk64CellMap );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                    Grid2DSquareCellDoubleChunkArray grid2DSquareCellDoubleChunkArray =
                            ( Grid2DSquareCellDoubleChunkArray ) object0;
                    grid2DSquareCellDoubleChunkArray.initGrid2DSquareCell( this );
                    grid2DSquareCellDoubleChunkArray.initChunkID( _ChunkID );
                    this._Grid2DSquareCellChunks.put(
                            _ChunkID,
                            grid2DSquareCellDoubleChunkArray );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                    Grid2DSquareCellDoubleChunkMap grid2DSquareCellDoubleChunkMap =
                            ( Grid2DSquareCellDoubleChunkMap ) object0;
                    grid2DSquareCellDoubleChunkMap.initGrid2DSquareCell( this );
                    grid2DSquareCellDoubleChunkMap.initChunkID( _ChunkID );
                    this._Grid2DSquareCellChunks.put(
                            _ChunkID,
                            grid2DSquareCellDoubleChunkMap );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                    Grid2DSquareCellDoubleChunkJAI grid2DSquareCellDoubleChunkJAI =
                            ( Grid2DSquareCellDoubleChunkJAI ) object0;
                    grid2DSquareCellDoubleChunkJAI.initGrid2DSquareCell( this );
                    grid2DSquareCellDoubleChunkJAI.initChunkID( _ChunkID );
                    this._Grid2DSquareCellChunks.put(
                            _ChunkID,
                            grid2DSquareCellDoubleChunkJAI );
                    return;
                }
                System.err.println(
                        "Unrecognised type of Grid2DSquareCellDoubleChunkAbstract or null " +
                        this.getClass().getName() +
                        ".loadIntoCacheChunk( ChunkID( " + _ChunkID.toString() + " ) )" );
            }
        }
    }
    
    /**
     * Loads an object from _File.
     * @param _File The file from which the Object is loaded.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The loaded Object or null.
     */
    protected final Object loadObject(
            File _File,
            boolean _HandleOutOfMemoryError ) {
        try {
            return loadObject( _File );
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return loadObject(
                        _File,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Loads an object from _File.
     * @param _File The file from which the Object is loaded.
     * @return The loaded Object or null.
     */
    protected final Object loadObject(
            File _File ) {
        //boolean _FileExists = _File.exists();
        //boolean _FileIsFile = _File.isFile();
        boolean _FileCanRead = _File.canRead();
        if ( ! _FileCanRead ) {
            System.err.println("Cannot read file " + _File + " in " + this.getClass().getName() + ".loadObject(File)");
            throw new Error();
        }
        FileInputStream _FileInputStream = null;
        BufferedInputStream _BufferedInputStream = null;
        ObjectInputStream _ObjectInputStream = null;
        try {
            _FileInputStream = new FileInputStream( _File );
        } catch ( java.io.FileNotFoundException _FileNotFoundException ) {
            System.err.println("" + _FileNotFoundException.toString() + " in " + this.getClass().getName() + ".loadObject(File)" );
        }
        //try {
            _BufferedInputStream = new BufferedInputStream( _FileInputStream );
        //} catch ( java.io.IOException _IOException ) {
        //    System.err.println("" + _IOException.toString() + " in " + this.getClass().getName() + ".loadObject(File)" );
        //}
        //_ObjectInputStream = new ObjectInputStream( _BufferedInputStream );
        try {
            _ObjectInputStream = new ObjectInputStream(
                    new BufferedInputStream(
                    new FileInputStream( _File ) ) );
        } catch ( Exception e0 ) {
            if ( _ObjectInputStream != null ) {
                try {
                    _ObjectInputStream.close();
                } catch ( Exception e1 ) {
                    e0.printStackTrace();
                }
            }
            //return null;
        } catch ( Error e0 ) {
            if ( _ObjectInputStream != null ) {
                try {
                    _ObjectInputStream.close();
                } catch ( Exception e1 ) {
                    e0.printStackTrace();
                }
            }
        }
        return loadObject( _ObjectInputStream );
    }
    
    /**
     * Loads an object from _ObjectInputStream.
     * @param _ObjectInputStream The ObjectInputStream from which the Object is loaded.
     * @return The loaded Object or null.
     */
    protected final Object loadObject(
            ObjectInputStream _ObjectInputStream ) {
        Object result = null;
        try {
            result = _ObjectInputStream.readObject();
            _ObjectInputStream.close();
        } catch ( Exception e0 ) {
            e0.printStackTrace();
        }
        return result;
    }
    
    /**
     * @return a CellID[] - the cell IDs for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     * returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     * returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID[] getCellIDs(
            double x,
            double y,
            double distance,
            boolean _HandleOutOfMemoryError ) {
        try {
            CellID[] result = getCellIDs(
                    x,
                    y,
                    distance );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellIDs(
                        x,
                        y,
                        distance,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a CellID[] - the cell IDs for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     * returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     * returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     */
    protected final CellID[] getCellIDs(
            double x,
            double y,
            double distance ) {
        return getCellIDs(
                x,
                y,
                getCellRowIndex( y ),
                getCellColIndex( x ),
                distance );
    }
    
    /**
     * @return a CellID[] - the cell IDs for cells thats centroids would be
     * intersected by circle with centre at centroid of cell given by cell row
     * index _CellRowIndex, cell column index _CellColIndex, and radius distance.
     * @param _CellRowIndex the row index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param _CellColIndex the column index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID[] getCellIDs(
            long _CellRowIndex,
            long _CellColIndex,
            double distance,
            boolean _HandleOutOfMemoryError ) {
        try {
            CellID[] result = getCellIDs(
                    _CellRowIndex,
                    _CellColIndex,
                    distance );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellIDs(
                        _CellRowIndex,
                        _CellColIndex,
                        distance,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return a CellID[] - the cell IDs for cells thats centroids would be
     * intersected by circle with centre at centroid of cell given by cell row
     * index _CellRowIndex, cell column index _CellColIndex, and radius distance.
     * @param _CellRowIndex the row index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param _CellColIndex the column index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     */
    protected final CellID[] getCellIDs(
            long _CellRowIndex,
            long _CellColIndex,
            double distance ) {
        return getCellIDs(
                getCellXDouble( _CellColIndex ),
                getCellYDouble( _CellRowIndex ),
                _CellRowIndex,
                _CellColIndex,
                distance );
    }
    
    /**
     * @return double[] cells - the values for cells thats centroids would be
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     *        returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     *        returned.
     * @param _CellRowIndex the row index at y.
     * @param _CellColIndex the col index at x.
     * @param distance the radius of the circle for which intersected cell
     *        values are returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID[] getCellIDs(
            double x,
            double y,
            long _CellRowIndex,
            long _CellColIndex,
            double distance,
            boolean _HandleOutOfMemoryError ) {
        try {
            CellID[] result = getCellIDs(
                    x,
                    y,
                    _CellRowIndex,
                    _CellColIndex,
                    distance );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellIDs(
                        x,
                        y,
                        _CellRowIndex,
                        _CellColIndex,
                        distance,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * TODO:
     * remove need for copy to new array.
     * @return double[] cells - the values for cells thats centroids would be
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     *        returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     *        returned.
     * @param _CellRowIndex the row index at y.
     * @param _CellColIndex the col index at x.
     * @param distance the radius of the circle for which intersected cell
     *        values are returned.
     */
    protected CellID[] getCellIDs(
            double x,
            double y,
            long _CellRowIndex,
            long _CellColIndex,
            double distance ) {
        CellID[] _CellIDs0;
        int cellDistance = ( int ) Math.ceil( distance / this._Dimensions[ 0 ].doubleValue() );
        int limit = ( ( 2 * cellDistance ) + 1 ) * ( ( 2 * cellDistance ) + 1 );
        _CellIDs0 = new CellID[ limit ];
        long row;
        long col;
        double thisX;
        double thisY;
        int count = 0;
        //if ( limit > 0 ) {
        for ( row = _CellRowIndex - cellDistance; row <= _CellRowIndex + cellDistance; row ++ ) {
            thisY = getCellYDouble( row );
            for ( col = _CellColIndex - cellDistance; col <= _CellColIndex + cellDistance; col ++ ) {
                thisX = getCellXDouble( col );
                if ( Utilities.distance( x, y, thisX, thisY ) <= distance ) {
                    _CellIDs0[ count ] = new CellID(
                            this._NCols,
                            row,
                            col );
                    //cellIDs0[ count ] = new CellID( row, col );
                    count ++;
                }
            }
        }
        CellID[] _CellIDs = new CellID[ count ];
        System.arraycopy( _CellIDs0, 0, _CellIDs, 0, count );
        return _CellIDs;
        //}
        //return _CellIDs0;
    }
    
    /**
     * @return Nearest cells _CellRowIndex and _CellColIndex as a long[]
     * from ID to point given by x-coordinate x, y-coordinate y.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID getNearestCellID(
            double x,
            double y,
            boolean _HandleOutOfMemoryError ) {
        try {
            CellID result = getNearestCellID(
                    x,
                    y );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getNearestCellID(
                        x,
                        y,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Nearest cells _CellRowIndex and _CellColIndex as a long[]
     * from ID to point given by x-coordinate x, y-coordinate y.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     */
    protected CellID getNearestCellID(
            double x,
            double y ) {
        return getNearestCellID(
                x,
                y, getCellRowIndex( y ),
                getCellColIndex( x ) );
    }
    
    /**
     * @return Nearest cells _CellRowIndex and _CellColIndex as a long[]
     * from ID to point given by cell row index _CellRowIndex, cell column index
     * _CellColIndex.
     * @param _CellRowIndex the row index from which nearest cell ID is returned.
     * @param _CellColIndex the column index from which nearest cell ID is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * return CellID[] as could be more than one nearest CellID
     */
    public CellID getNearestCellID(
            long _CellRowIndex,
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            CellID result = getNearestCellID(
                    _CellRowIndex,
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getNearestCellID(
                        _CellRowIndex,
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Nearest cells _CellRowIndex and _CellColIndex as a long[]
     * from ID to point given by cell row index _CellRowIndex, cell column index
     * _CellColIndex.
     * @param _CellRowIndex the row index from which nearest cell ID is returned.
     * @param _CellColIndex the column index from which nearest cell ID is returned.
     * TODO:
     * return CellID[] as could be more than one nearest CellID
     */
    protected CellID getNearestCellID(
            long _CellRowIndex,
            long _CellColIndex ) {
        return getNearestCellID(
                getCellXDouble( _CellColIndex ),
                getCellYDouble( _CellRowIndex ),
                _CellRowIndex,
                _CellColIndex );
    }
    
    /**
     * @return Nearest CellID to point given by x-coordinate x,
     * y-coordinate y in position given by _CellRowIndex, _CellColIndex.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param _CellRowIndex The cell row index of cell containing point.
     * @param _CellColIndex The cell column index of cell containing point.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID getNearestCellID(
            double x,
            double y,
            long _CellRowIndex,
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            CellID result = getNearestCellID(
                    x,
                    y,
                    _CellRowIndex,
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getNearestCellID(
                        x,
                        y,
                        _CellRowIndex,
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Nearest CellID to point given by x-coordinate x,
     * y-coordinate y in position given by _CellRowIndex, _CellColIndex.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param _CellRowIndex The cell row index of cell containing point.
     * @param _CellColIndex The cell column index of cell containing point.
     */
    protected CellID getNearestCellID(
            double x,
            double y,
            long _CellRowIndex,
            long _CellColIndex ) {
        CellID _CellID;
        boolean isInGrid = isInGrid(
                x,
                y );
        if ( ! isInGrid ) {
            long row;
            long col;
            if ( x >= this._Dimensions[3].doubleValue() ) {
                col = this._NCols - 1;
                if ( y > this._Dimensions[4].doubleValue() ) {
                    row = 0;
                } else {
                    if ( y < this._Dimensions[2].doubleValue() ) {
                        row = this._NRows - 1;
                    } else {
                        row = getCellRowIndex( y );
                    }
                }
            } else {
                if ( x < this._Dimensions[1].doubleValue() ) {
                    col = 0;
                    if ( y >= this._Dimensions[4].doubleValue() ) {
                        row = 0;
                    } else {
                        if ( y < this._Dimensions[2].doubleValue() ) {
                            row = this._NRows -1;
                        } else {
                            row = getCellRowIndex( y );
                        }
                    }
                } else {
                    col = getCellColIndex( x );
                    if ( y >= _Dimensions[4].doubleValue() ) {
                        row = 0;
                    } else {
                        row = this._NRows - 1;
                    }
                }
            }
            _CellID = new CellID(
                    this._NCols,
                    row,
                    col );
        } else {
            _CellID = new CellID(
                    this._NCols,
                    _CellRowIndex,
                    _CellColIndex );
        }
        return _CellID;
    }
    
    /**
     * @return Height of the grid.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getHeightDouble(
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getHeightDouble();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getHeightDouble(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Height of the grid.
     */
    protected final double getHeightDouble() {
        return getHeightBigDecimal().doubleValue();
    }
    
    /**
     * @return Height of the grid.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getHeightBigDecimal(
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getHeightBigDecimal();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getHeightBigDecimal(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Height of the grid.
     */
    protected final BigDecimal getHeightBigDecimal() {
        return ( this._Dimensions[ 4 ].subtract( this._Dimensions[ 2 ] ) );
    }
    
    /**
     * @return Width of the grid as a double.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getWidthDouble(
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getWidthDouble();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getWidthDouble(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Width of the grid as a double.
     */
    protected final double getWidthDouble() {
        return getWidthBigDecimal().doubleValue();
    }
    
    /**
     * @return Width of the grid as a BigDecimal.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getWidthBigDecimal(
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getWidthBigDecimal();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getWidthBigDecimal(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return Width of the grid as a BigDecimal.
     */
    protected final BigDecimal getWidthBigDecimal() {
        return ( this._Dimensions[ 3 ].subtract( this._Dimensions[ 1 ] ) );
    }
    
    /**
     * @return true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param xBigDecimal The x-coordinate of the point.
     * @param yBigDecimal The y-coordinate of the point.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            BigDecimal xBigDecimal,
            BigDecimal yBigDecimal,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = isInGrid(
                    xBigDecimal,
                    yBigDecimal );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return isInGrid(
                        xBigDecimal,
                        yBigDecimal,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param xBigDecimal The x-coordinate of the point.
     * @param yBigDecimal The y-coordinate of the point.
     */
    protected final boolean isInGrid(
            BigDecimal xBigDecimal,
            BigDecimal yBigDecimal ) {
        return ( xBigDecimal.compareTo( this._Dimensions[ 1 ] ) != -1 &&
                yBigDecimal.compareTo( this._Dimensions[ 2 ] ) != -1 &&
                xBigDecimal.compareTo( this._Dimensions[ 3 ] ) != 1 &&
                yBigDecimal.compareTo( this._Dimensions[ 4 ] ) != 1 );
    }
    
    /**
     * @return true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            double x,
            double y,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = isInGrid(
                    x,
                    y );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return isInGrid(
                        x,
                        y,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     */
    protected final boolean isInGrid(
            double x,
            double y ) {
        return isInGrid(
                BigDecimal.valueOf( x ),
                BigDecimal.valueOf( y ) );
    }
    
    /**
     * @return true iff position given by cell row index _CellRowIndex, cell
     * column index _CellColIndex is in the Grid.
     * @param _CellRowIndex The cell row index to test.
     * @param _CellColIndex The cell column index to test.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            long _CellRowIndex,
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = isInGrid(
                    _CellRowIndex,
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return isInGrid(
                        _CellRowIndex,
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return true iff position given by cell row index _CellRowIndex, cell
     * column index _CellColIndex is in the Grid.
     * @param _CellRowIndex The cell row index to test.
     * @param _CellColIndex The cell column index to test.
     */
    protected final boolean isInGrid(
            long _CellRowIndex,
            long _CellColIndex ) {
        return ( _CellRowIndex >= 0 &&
                _CellRowIndex < this._NRows &&
                _CellColIndex >= 0 &&
                _CellColIndex < this._NCols );
    }
    
    /**
     * @return true iff position given by chunk row index _ChunkRowIndex, chunk
     * column index _ChunkColIndex is in the Grid.
     * @param _ChunkRowIndex The chunk row index to test.
     * @param _ChunkColIndex The chunk column index to test.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = isInGrid(
                    _ChunkRowIndex,
                    _ChunkColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return isInGrid(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return true iff position given by chunk row index _ChunkRowIndex, chunk
     * column index _ChunkColIndex is in the Grid.
     *
     *
     * @param _ChunkRowIndex The chunk row index to test.
     * @param _ChunkColIndex The chunk column index to test.
     */
    protected final boolean isInGrid(
            int _ChunkRowIndex,
            int _ChunkColIndex ) {
        return ( _ChunkRowIndex >= 0 &&
                _ChunkRowIndex < this._NRows &&
                _ChunkColIndex >= 0 &&
                _ChunkColIndex < this._NCols );
    }
    
    /**
     * @return true iff cell given by _CellID is in the Grid.
     * @param _CellID The CellID of a cell to test.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            CellID _CellID,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = isInGrid( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return isInGrid(
                        _CellID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return true iff cell given by _CellID is in the Grid.
     * @param _CellID The CellID of a cell to test.
     */
    protected final boolean isInGrid(
            CellID _CellID ) {
        return isInGrid(
                _CellID.getCellRowIndex(),
                _CellID.getCellColIndex() );
    }
    
    /**
     * @return true iff ChunkID _ChunkID is in the Grid.
     * @param _ChunkID The ChunkID of a cell to test.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            ChunkID _ChunkID,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = isInGrid( _ChunkID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return isInGrid(
                        _ChunkID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return true iff ChunkID _ChunkID is in the Grid.
     * @param _ChunkID The ChunkID of a cell to test.
     */
    protected final boolean isInGrid(
            ChunkID _ChunkID ) {
        int _ChunkRowIndex = _ChunkID._ChunkRowIndex;
        int _ChunkColIndex = _ChunkID._ChunkColIndex;
        return _ChunkRowIndex > -1 &&
                _ChunkRowIndex < this._NChunkRows &&
                _ChunkColIndex > -1 &&
                _ChunkColIndex < this._NChunkCols;
    }
    
    /**
     * @return true iff cell given by _ChunkRowIndex, _ChunkColIndex,
     * chunkCellRowIndex, chunkCellColIndex is in the Grid.
     * @param _ChunkRowIndex
     * @param _ChunkColIndex
     * @param chunkCellRowIndex
     * @param chunkCellColIndex
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            boolean result = isInGrid(
                    _ChunkRowIndex,
                    _ChunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return isInGrid(
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return true iff cell given by _ChunkRowIndex, _ChunkColIndex,
     * chunkCellRowIndex, chunkCellColIndex is in the Grid.
     * @param _ChunkRowIndex
     * @param _ChunkColIndex
     * @param chunkCellRowIndex
     * @param chunkCellColIndex
     */
    protected final boolean isInGrid(
            int _ChunkRowIndex,
            int _ChunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        return isInGrid(
                ( ( long ) _ChunkRowIndex * ( long ) this._ChunkNCols ) + ( long ) chunkCellRowIndex,
                ( ( long ) _ChunkColIndex * ( long ) this._ChunkNRows ) + ( long ) chunkCellColIndex );
    }
    
    /**
     * @return the x-coordinate of the centroid for cells with column index
     * _CellColIndex as a BigDecimal.
     * @param _CellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellXBigDecimal(
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getCellXBigDecimal( _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellXBigDecimal(
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return x-coordinate of the centroid for cells with column index
     * _CellColIndex as a BigDecimal.
     * @param _CellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellXBigDecimal(
            long _CellColIndex,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getCellXBigDecimal( _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError  );
                return getCellXBigDecimal(
                        _CellColIndex,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the x-coordinate of the centroid for cells with column index
     * _CellColIndex as a BigDecimal.
     * @param _CellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     */
    protected final BigDecimal getCellXBigDecimal(
            long _CellColIndex ) {
        BigDecimal _BigDecimal0 = BigDecimal.valueOf( _CellColIndex );
        BigDecimal _BigDecimal1 = _Dimensions[ 0 ].multiply( _BigDecimal0 );
        BigDecimal _BigDecimal2 = _Dimensions[ 1 ].add( _BigDecimal1 );
        _BigDecimal1 =  BigDecimal.valueOf( 2L );
        _BigDecimal0 = _Dimensions[ 0 ].divide(
                _BigDecimal1,
                this._DimensionsScale,
                BigDecimal.ROUND_HALF_EVEN );
        return _BigDecimal2.add( _BigDecimal0 );
    }
    
    /**
     * @return x-coordinate of the centroid for cells with column index
     * _CellColIndex as a double.
     * @param _CellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellXDouble(
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellXDouble(
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellXDouble(
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return x-coordinate of the centroid for cells with column index
     * _CellColIndex as a double.
     * @param _CellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param _ChunkID This is a ChunkID for those AbstractGrid2DSquareCells not
     * to be swapped if possible when an OutOfMemoryError is encountered.
     */
    public final double getCellXDouble(
            long _CellColIndex,
            boolean _HandleOutOfMemoryError,
            ChunkID _ChunkID ) {
        try {
            double result = getCellXDouble(
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _SwapToFileGrid2DSquareCellChunkExcept( _ChunkID ) == null ) {
                    _SwapToFileGrid2DSquareCellChunk();
                }
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellXDouble(
                        _CellColIndex,
                        _HandleOutOfMemoryError,
                        _ChunkID );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return x-coordinate of the centroid for cells with column index
     * _CellColIndex as a double.
     * @param chunkCellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param _ChunkRowIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellXDouble(
            int chunkCellColIndex,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellXDouble(
                    chunkCellColIndex,
                    _ChunkColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellXDouble(
                        chunkCellColIndex,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return x-coordinate of the centroid for cells with column index
     * _CellColIndex as a double.
     * @param chunkCellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     */
    protected final double getCellXDouble(
            int chunkCellColIndex,
            int _ChunkColIndex ) {
        return getCellXDouble( getCellColIndex( chunkCellColIndex, _ChunkColIndex ) );
    }
    
    /**
     * @return x-coordinate of the centroid for cells with column index
     * _CellColIndex as a double.
     * @param _CellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     */
    protected final double getCellXDouble(
            long _CellColIndex ) {
        return getCellXBigDecimal( _CellColIndex ).doubleValue();
    }
    
    /**
     * @return x-coordinate of the centroid for cell with cell ID _CellID as
     * a BigDecimal.
     * @param _CellID
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellXBigDecimal(
            CellID _CellID,
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getCellXBigDecimal( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellXBigDecimal(
                        _CellID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @param _CellID The CellID of the cell thats centroid is returned.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @return x-coordinate of the centroid of cell with CellID _CellID as a
     * getCellXBigDecimal.
     */
    public final BigDecimal getCellXBigDecimal(
            CellID _CellID,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getCellXBigDecimal( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellXBigDecimal(
                        _CellID,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return x-coordinate of the centroid for cell with cell ID _CellID as
     * a BigDecimal.
     * @param _CellID
     */
    protected final BigDecimal getCellXBigDecimal(
            CellID _CellID ) {
        BigDecimal _BigDecimal0 = BigDecimal.valueOf( _CellID._CellColIndex );
        BigDecimal _BigDecimal1 = _Dimensions[ 0 ].multiply( _BigDecimal0 );
        BigDecimal _BigDecimal2 = _Dimensions[ 1 ].add( _BigDecimal1 );
        _BigDecimal0 = BigDecimal.valueOf( 2L );
        _BigDecimal1 = _Dimensions[ 0 ].divide(
                _BigDecimal0,
                this._DimensionsScale,
                BigDecimal.ROUND_HALF_EVEN );
        return _BigDecimal2.add( _BigDecimal1 );
    }
    
    /**
     * @return x-coordinate of the centroid for cell with cell ID _CellID as
     * a double
     * @param _CellID
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellXDouble(
            CellID _CellID,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellXDouble( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellXDouble(
                        _CellID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return x-coordinate of the centroid of cell with CellID _CellID as a
     * double.
     * @param _CellID The CellID of the cell thats centroid is returned.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     */
    public final double getCellXDouble(
            CellID _CellID,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellXDouble( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellXDouble(
                        _CellID,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return x-coordinate of the centroid for cell with cell ID _CellID as
     * a double
     * @param _CellID
     */
    protected final double getCellXDouble(
            CellID _CellID ) {
        return getCellXBigDecimal( _CellID ).doubleValue();
    }
    
    /**
     * @return y-coordinate of the centroid for cells with row index
     * _CellRowIndex as a BigDecimal.
     * @param _CellRowIndex the cell column index thats centroid y-coordinate is
     *   returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellYBigDecimal(
            long _CellRowIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getCellYBigDecimal( _CellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYBigDecimal(
                        _CellRowIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return y-coordinate of the centroid for cells with row index
     * _CellRowIndex as a BigDecimal.
     * @param _CellRowIndex the cell column index thats centroid y-coordinate is
     *   returned.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     */
    public final BigDecimal getCellYBigDecimal(
            long _CellRowIndex,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getCellYBigDecimal( _CellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYBigDecimal(
                        _CellRowIndex,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return y-coordinate of the centroid for cells with row index
     * _CellRowIndex as a BigDecimal.
     * @param _CellRowIndex the cell column index thats centroid y-coordinate is
     *   returned.
     */
    protected final BigDecimal getCellYBigDecimal(
            long _CellRowIndex ) {
        BigDecimal _BigDecimal0 = BigDecimal.valueOf( _CellRowIndex );
        BigDecimal _BigDecimal1 = _Dimensions[ 0 ].multiply( _BigDecimal0 );
        BigDecimal _BigDecimal2 = _Dimensions[ 2 ].add( _BigDecimal1 );
        _BigDecimal0 = BigDecimal.valueOf( 2L );
        _BigDecimal1 = _Dimensions[ 0 ].divide(
                _BigDecimal0,
                this._DimensionsScale,
                BigDecimal.ROUND_HALF_EVEN );
        return _BigDecimal2.add( _BigDecimal1 );
    }
    
    /**
     * @return y-coordinate of the centroid for cells with row index
     * _CellRowIndex as a double.
     * @param _CellRowIndex the cell column index thats centroid y-coordinate is
     * returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellYDouble(
            long _CellRowIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellYDouble( _CellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYDouble(
                        _CellRowIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return y-coordinate of the centroid for cells with row index
     * _CellRowIndex as a double.
     * @param _CellRowIndex the cell column index thats centroid y-coordinate is
     * returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param _ChunkID This is a ChunkID for those AbstractGrid2DSquareCells not
     * to be swapped if possible when an OutOfMemoryError is encountered.
     */
    public final double getCellYDouble(
            long _CellRowIndex,
            boolean _HandleOutOfMemoryError,
            ChunkID _ChunkID ) {
        try {
            double result = getCellYDouble( _CellRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                if ( _SwapToFileGrid2DSquareCellChunkExcept( _ChunkID ) == null ) {
                    _SwapToFileGrid2DSquareCellChunk();
                }
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYDouble(
                        _CellRowIndex,
                        _HandleOutOfMemoryError,
                        _ChunkID );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return y-coordinate of the centroid for cells with row index
     * _CellRowIndex as a double.
     * @param chunkCellRowIndex the chunk cell column index thats centroid
     *   y-coordinate is returned.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellYDouble(
            int chunkCellRowIndex,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellYDouble( chunkCellRowIndex, _ChunkRowIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYDouble(
                        chunkCellRowIndex,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return y-coordinate of the centroid for cells with row index
     * _CellRowIndex as a double.
     * @param chunkCellRowIndex the chunk cell column index thats centroid
     *   y-coordinate is returned.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     */
    protected final double getCellYDouble(
            int chunkCellRowIndex,
            int _ChunkRowIndex ) {
        return getCellYDouble( getCellRowIndex( chunkCellRowIndex, _ChunkRowIndex ) );
    }
    
    /**
     * @return y-coordinate of the centroid for cells with row index
     * _CellRowIndex as a double.
     * @param _CellRowIndex the cell column index thats centroid y-coordinate is
     * returned.
     */
    protected final double getCellYDouble(
            long _CellRowIndex ) {
        return getCellYBigDecimal( _CellRowIndex ).doubleValue();
    }
    
    /**
     * @return y-coordinate of the centroid of cell with CellID _CellID as a
     * BigDecimal.
     * @param _CellID The CellID of the cell thats centroid is returned.
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellYBigDecimal(
            CellID _CellID,
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getCellYBigDecimal( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYBigDecimal(
                        _CellID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @param _CellID The CellID of the cell thats centroid is returned.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @return y-coordinate of the centroid of cell with CellID _CellID as a
     * BigDecimal.
     */
    public final BigDecimal getCellYBigDecimal(
            CellID _CellID,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            BigDecimal result = getCellYBigDecimal( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYBigDecimal(
                        _CellID,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return y-coordinate of the centroid of cell with CellID _CellID as a
     * BigDecimal.
     * @param _CellID The CellID of the cell thats centroid is returned.
     */
    protected final BigDecimal getCellYBigDecimal(
            CellID _CellID ) {
        BigDecimal _BigDecimal0 = BigDecimal.valueOf( _CellID._CellRowIndex );
        BigDecimal _BigDecimal1 = _Dimensions[ 0 ].multiply( _BigDecimal0 );
        BigDecimal _BigDecimal2 = _Dimensions[ 2 ].add( _BigDecimal1 );
        _BigDecimal0 = BigDecimal.valueOf( 2L );
        _BigDecimal1 = _Dimensions[ 0 ].divide(
                _BigDecimal0,
                this._DimensionsScale,
                BigDecimal.ROUND_HALF_EVEN );
        return _BigDecimal2.add( _BigDecimal1 );
    }
    
    /**
     * @return y-coordinate of the centroid of cell with CellID _CellID as a
     * double.
     * @param _CellID
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellYDouble(
            CellID _CellID,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellYDouble( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYDouble(
                        _CellID,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return y-coordinate of the centroid of cell with CellID _CellID as a
     * double.
     * @param _CellID The CellID of the cell thats centroid is returned.
     * @param _ChunkRowIndex The chunk row index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     * @param _ChunkColIndex The chunk column index of the
     *   AbstractGrid2DSquareCellChunk not to be swapped if an OutOfMemoryError
     *   is thrown.
     */
    public final double getCellYDouble(
            CellID _CellID,
            int _ChunkRowIndex,
            int _ChunkColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double result = getCellYDouble( _CellID );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellYDouble(
                        _CellID,
                        _ChunkRowIndex,
                        _ChunkColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the y-coordinate of the centroid of cell with CellID _CellID as a
     * double.
     * @param _CellID
     */
    protected final double getCellYDouble(
            CellID _CellID ) {
        return getCellYBigDecimal( _CellID ).doubleValue();
    }
    
    /**
     * TODO:
     * Are bounds in double range?
     * Is there more than cellsize difference with precision?
     *   Throw appropriate exceptions.
     * @return gridBounds (the bounding box of the grid) as a double[] where;
     * gridBounds[0] xmin, left most x-coordinate of this
     * gridBounds[1] ymin, lowest y-coordinate of this
     * gridBounds[2] xmax, right most x-coordinate of this
     * gridBounds[3] ymax, highest y-coordinate of this
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double[] getGridBounds(
            boolean _HandleOutOfMemoryError ) {
        try {
            double[] result = getGridBounds();
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getGridBounds(
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return gridBounds (the bounding box of the grid) as a double[] where;
     * gridBounds[0] xmin, left most x-coordinate of this
     * gridBounds[1] ymin, lowest y-coordinate of this
     * gridBounds[2] xmax, right most x-coordinate of this
     * gridBounds[3] ymax, highest y-coordinate of this
     * TODO:
     * Are bounds in double range?
     * Is there more than cellsize difference with precision?
     *   Throw appropriate exceptions.
     */
    protected final double[] getGridBounds() {
        double[] result = new double[ 4 ];
        result[ 0 ] = _Dimensions[ 1 ].doubleValue();
        result[ 1 ] = _Dimensions[ 2 ].doubleValue();
        result[ 2 ] = _Dimensions[ 3 ].doubleValue();
        result[ 3 ] = _Dimensions[ 4 ].doubleValue();
        return result;
    }
    
    /**
     * @return double[] where;
     * double[0] xmin, left most x-coordinate of cell at (rowIndex,colIndex)
     * double[1] ymin, lowest y-coordinate of cell at (rowIndex,colIndex)
     * double[2] xmax, right most x-coordinate of cell at (rowIndex,colIndex)
     * double[3] ymax, highest y-coordinate of cell at (rowIndex,colIndex)
     * @param _CellRowIndex the row index of the cell for which the bounds are
     *   returned
     * @param _CellColIndex the column index of the cell for which the bounds are
     *   returned
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double[] getCellBoundsDoubleArray(
            long _CellRowIndex,
            long _CellColIndex,
            boolean _HandleOutOfMemoryError ) {
        try {
            double[] result = getCellBoundsDoubleArray(
                    _CellRowIndex,
                    _CellColIndex );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellBoundsDoubleArray(
                        _CellRowIndex,
                        _CellColIndex,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return double[] where;
     * double[0] xmin, left most x-coordinate of cell at (rowIndex,colIndex)
     * double[1] ymin, lowest y-coordinate of cell at (rowIndex,colIndex)
     * double[2] xmax, right most x-coordinate of cell at (rowIndex,colIndex)
     * double[3] ymax, highest y-coordinate of cell at (rowIndex,colIndex)
     * @param _CellRowIndex the row index of the cell for which the bounds are
     *   returned
     * @param _CellColIndex the column index of the cell for which the bounds are
     *   returned
     */
    protected final double[] getCellBoundsDoubleArray(
            long _CellRowIndex,
            long _CellColIndex ) {
        return getCellBoundsDoubleArray(
                getCellXDouble( _CellColIndex ),
                getCellYDouble( _CellRowIndex ) );
    }
    
    /**
     * Precision may compromise result. More precision is available via
     * @return double[] where;
     * double[0] xmin, left most x-coordinate of cell that intersects point at (x,y)
     * double[1] ymin, lowest y-coordinate of cell that intersects point at (x,y)
     * double[2] xmax, right most x-coordinate of cell that intersects point at (x,y)
     * double[3] ymax, highest y-coordinate of cell that intersects point at (x,y)
     * @param x the x-coordinate in the cell for which the bounds are returned
     * @param y the y-coordinate in the cell for which the bounds are returned
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double[] getCellBoundsDoubleArray(
            double x,
            double y,
            boolean _HandleOutOfMemoryError ) {
        try {
            double[] result = getCellBoundsDoubleArray(
                    x,
                    y );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getCellBoundsDoubleArray(
                        x,
                        y,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * Precision may compromise result. More precision is available via
     * BigDecimal arithmetic.
     * @return double[] where;
     * double[0] xmin, left most x-coordinate of cell that intersects point at (x,y)
     * double[1] ymin, lowest y-coordinate of cell that intersects point at (x,y)
     * double[2] xmax, right most x-coordinate of cell that intersects point at (x,y)
     * double[3] ymax, highest y-coordinate of cell that intersects point at (x,y)
     * @param x the x-coordinate in the cell for which the bounds are returned
     * @param y the y-coordinate in the cell for which the bounds are returned
     */
    protected final double[] getCellBoundsDoubleArray(
            double x,
            double y ) {
        double[] cellBounds = new double[ 4 ];
        double halfCellsize = _Dimensions[ 0 ].doubleValue() / 2.0d;
        cellBounds[ 0 ] = x - halfCellsize;
        cellBounds[ 1 ] = y - halfCellsize;
        cellBounds[ 2 ] = x + halfCellsize;
        cellBounds[ 3 ] = y + halfCellsize;
        return cellBounds;
    }
    
    /**
     * @return the _NoDataValue of this as a BigDecimal.
     */
    public abstract BigDecimal getNoDataValueBigDecimal(
            boolean _HandleOutOfMemoryError );
    
    /**
     * @return the next ChunkID in row major order from _ChunkID, or null.
     * @param _ChunkID
     * @param _NChunkRows
     * @param _NChunkCols
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public ChunkID getNextChunk(
            ChunkID _ChunkID,
            int _NChunkRows,
            int _NChunkCols,
            boolean _HandleOutOfMemoryError ) {
        try {
            ChunkID result = getNextChunk(
                    _ChunkID,
                    _NChunkRows,
                    _NChunkCols );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return  getNextChunk(
                        _ChunkID,
                        _NChunkRows,
                        _NChunkCols,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the next ChunkID in row major order from _ChunkID, or null.
     * @param _ChunkID
     * @param _NChunkRows
     * @param _NChunkCols
     */
    protected ChunkID getNextChunk(
            ChunkID _ChunkID,
            int _NChunkRows,
            int _NChunkCols ) {
        int _ChunkRowIndex = _ChunkID.getChunkRowIndex();
        int _ChunkColIndex = _ChunkID.getChunkColIndex();
        if ( _ChunkColIndex < _NChunkCols - 1 ) {
            return new ChunkID(
                    _NChunkCols,
                    _ChunkRowIndex,
                    _ChunkColIndex + 1 );
        } else {
            if ( _ChunkRowIndex < _NChunkRows - 1  ) {
                return new ChunkID(
                        _NChunkCols,
                        _ChunkRowIndex + 1,
                        0 );
            }
        }
        return null;
    }
    
    /**
     * @return the next ChunkID in row minor order from this, or null.
     * @param _ChunkID
     * @param _NChunkRows
     * @param _NChunkCols
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public ChunkID getNextChunkRowMinor(
            ChunkID _ChunkID,
            int _NChunkRows,
            int _NChunkCols,
            boolean _HandleOutOfMemoryError ) {
        try {
            ChunkID result = getNextChunkRowMinor(
                    _ChunkID,
                    _NChunkRows,
                    _NChunkCols );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getNextChunkRowMinor(
                        _ChunkID,
                        _NChunkRows,
                        _NChunkCols,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the next ChunkID in row minor order from this, or null.
     * @param _ChunkID
     * @param _NChunkRows
     * @param _NChunkCols
     */
    protected ChunkID getNextChunkRowMinor(
            ChunkID _ChunkID,
            int _NChunkRows,
            int _NChunkCols ) {
        int _ChunkRowIndex = _ChunkID.getChunkRowIndex();
        int _ChunkColIndex = _ChunkID.getChunkColIndex();
        if ( _ChunkColIndex < _NChunkCols - 1 ) {
            return new ChunkID(
                    _NChunkCols,
                    _ChunkRowIndex,
                    _ChunkColIndex + 1 );
        } else {
            if ( _ChunkRowIndex > 0 ) {
                return new ChunkID(
                        _NChunkCols,
                        _ChunkRowIndex - 1,
                        0 );
            }
        }
        return null;
    }
    
    /**
     * @return the next ChunkID in row major order from this, or null.
     * @param _ChunkID
     * @param _NChunkRows
     * @param _NChunkCols
     * @param _HandleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public ChunkID getPreviousChunk(
            ChunkID _ChunkID,
            int _NChunkRows,
            int _NChunkCols,
            boolean _HandleOutOfMemoryError ) {
        try {
            ChunkID result = getPreviousChunk(
                    _ChunkID,
                    _NChunkRows,
                    _NChunkCols );
            _EnsureThereIsEnoughMemoryToContinue();
            return result;
        } catch ( OutOfMemoryError _OutOfMemoryError ) {
            if ( _HandleOutOfMemoryError ) {
                clear_MemoryReserve();
                _SwapToFileGrid2DSquareCellChunk();
                init_MemoryReserve( _HandleOutOfMemoryError );
                return getPreviousChunk(
                        _ChunkID,
                        _NChunkRows,
                        _NChunkCols,
                        _HandleOutOfMemoryError );
            } else {
                throw _OutOfMemoryError;
            }
        }
    }
    
    /**
     * @return the next ChunkID in row major order from this, or null.
     *
     *
     * @param _ChunkID
     * @param _NChunkRows
     * @param _NChunkCols
     */
    protected ChunkID getPreviousChunk(
            ChunkID _ChunkID,
            int _NChunkRows,
            int _NChunkCols ) {
        int _ChunkRowIndex = _ChunkID.getChunkRowIndex();
        int _ChunkColIndex = _ChunkID.getChunkColIndex();
        if ( _ChunkColIndex > 0 ) {
            return new ChunkID(
                    _NChunkCols,
                    _ChunkRowIndex,
                    _ChunkColIndex - 1 );
        } else {
            if ( _ChunkRowIndex > 0  ) {
                return new ChunkID(
                        _NChunkCols,
                        _ChunkRowIndex - 1,
                        0 );
            }
        }
        return null;
    }
    
    
    /**
     * A simple CellID class for distinguishing cells.
     */
    public static class CellID extends Object implements Serializable, Comparable {
        
        /**
         * For storing the row in which the cell is positioned
         */
        protected long _CellRowIndex;
        
        /**
         * For storing the column in which the cell is positioned
         */
        protected long _CellColIndex;
        
        /**
         * For storing a hashCode for the cell.
         * NB. When the number of cells in a grid is more than the maximum int
         * value care should be taken that hashCodes are distributed in a
         * relatively even way. Operations will slow down greatly if a large
         * number of cells have a CellID with the same hashCode.
         */
        protected int hashCode;
        
        /**
         * Default constructor.
         */
        public CellID() {}
        
        /**
         * Constructs a new CellID that is a clone of _CellID
         * @param _CellID The CellID to clone.
         */
        public CellID( CellID _CellID ) {
            this._CellColIndex = _CellID._CellColIndex;
            this._CellRowIndex = _CellID._CellRowIndex;
            this.hashCode = _CellID.hashCode;
        }
        
        /**
         * Constructs a new CellID.
         *
         * @param _CellRowIndex The row of the cell.
         * @param _CellColIndex The column of the cell.
         * @param _NCols The number of columns in a grid2DSquareCellAbstract.
         */
        public CellID(
                long _NCols,
                long _CellRowIndex,
                long _CellColIndex ) {
            //this( _NRows, _NCols, _CellRowIndex, _CellColIndex );
            this._CellRowIndex = _CellRowIndex;
            this._CellColIndex = _CellColIndex;
            this.hashCode = ( int ) ( ( _NCols * _CellRowIndex ) + _CellColIndex );
        }
        
        //public CellID( long _CellRowIndex, long _CellColIndex ) {
        //    this( _NCols, _CellRowIndex, _CellColIndex );
        //    this._CellRowIndex = _CellRowIndex;
        //    this._CellColIndex = _CellColIndex;
        //    this.hashCode = ( int ) ( ( this._NCols * _CellRowIndex ) + _CellColIndex );
        //}
        
        /**
         * @return hashCode of this CellID
         */
        public int hashCode() {
            return this.hashCode;
        }
        
        /**
         * @return _CellRowIndex of this CellID
         */
        public long getCellRowIndex() {
            return this._CellRowIndex;
        }
        
        /**
         * @return _CellColIndex of this CellID
         */
        public long getCellColIndex() {
            return this._CellColIndex;
        }
        
        /**
         * @return a description of this CellID
         */
        public String toString() {
            return "CellID( hashCode( " + hashCode() + "), _CellRowIndex( " + getCellRowIndex() + " ), _CellColIndex( " + getCellColIndex() + " ) )";
        }
        
        /**
         * Overrides equals in Object
         */
        public boolean equals( Object object ) {
            if ( this == object ) {
                return true;
            }
            if ( ( object == null ) || ( object.getClass() != this.getClass() ) ) {
                return false;
            }
            CellID _CellID = ( CellID ) object;
            return ( ( this._CellColIndex == _CellID._CellColIndex ) && ( this._CellRowIndex == _CellID._CellRowIndex ) );
        }
        
        /**
         * Method required by Comparable
         */
        public int compareTo( Object object ) {
            //try {
            // Cast failure should result in NullPointerException or ClassCastException
            CellID _CellID = ( CellID ) object;
            if ( _CellID._CellRowIndex > this._CellRowIndex ) {
                return 1;
            }
            if ( _CellID._CellRowIndex < this._CellRowIndex ) {
                return -1;
            }
            if ( _CellID._CellColIndex > this._CellColIndex ) {
                return 1;
            }
            if ( _CellID._CellColIndex < this._CellColIndex ) {
                return -1;
            }
            return 0;
            //} catch ( Exception e0 ) {
            //    return -1;
            //}
        }
    }
    
    /**
     * A simple ChunkID class for distinguishing chunks.
     */
    public static class ChunkID extends Object implements Serializable, Comparable {
        
        /**
         * For storing the chunk row index
         */
        protected int _ChunkRowIndex;
        
        /**
         * For storing the chunk column index
         */
        protected int _ChunkColIndex;
        
        /**
         * For storing a hashCode for this ChunkID.
         * NB. When the number of cells in a chunk is more than the maximum int
         * value care should be taken to distributed hashcodes in a relatively
         * even way. Operations may slow down greatly if a large number of
         * chunks have ChunkIDs with the same hashCode.
         */
        protected int hashCode;
        
        /**
         * Default constructor.
         */
        public ChunkID() {
            this._ChunkColIndex = Integer.MIN_VALUE;
            this._ChunkRowIndex = Integer.MIN_VALUE;
            this.hashCode = Integer.MIN_VALUE;
        }
        
        /**
         * Constructs a new ChunkID that is a clone of _ChunkID
         * @param _ChunkID The ChunkID to clone.
         */
        public ChunkID( ChunkID _ChunkID ) {
            this._ChunkColIndex = _ChunkID._ChunkColIndex;
            this._ChunkRowIndex = _ChunkID._ChunkRowIndex;
            this.hashCode = _ChunkID.hashCode;
        }
        
        /**
         * Constructs a new ChunkID
         * @param _ChunkRowIndex the row of the
         * @param _ChunkColIndex the column of the
         * @param _NChunkCols The number of chunk columns (of the grid2DSquareCellAbstract).
         */
        public ChunkID(
                int _NChunkCols,
                int _ChunkRowIndex,
                int _ChunkColIndex ) {
            this._ChunkRowIndex = _ChunkRowIndex;
            this._ChunkColIndex = _ChunkColIndex;
            this.hashCode = ( int ) ( ( _NChunkCols * _ChunkRowIndex ) + _ChunkColIndex );
        }
        
        /**
         * @return this.hashCode
         */
        public int hashCode() {
            return this.hashCode;
        }
        
        /**
         * @return this._ChunkRowIndex
         */
        public int getChunkRowIndex() {
            return this._ChunkRowIndex;
        }
        
        /**
         * @return this._ChunkColIndex
         */
        public int getChunkColIndex() {
            return this._ChunkColIndex;
        }
        
        /**
         * @return a description of this
         */
        public String toString() {
            return  "ChunkID( " +
                    "hashCode( " + hashCode() + "), " +
                    "_ChunkRowIndex( " + getChunkRowIndex() + " ), " +
                    "_ChunkColIndex( " + getChunkColIndex() + " ) )";
        }
        
        /**
         * Overrides equals in Object
         */
        public boolean equals( Object object ) {
            if ( this == object ) {
                return true;
            }
            if ( ( object == null ) || ( object.getClass() != this.getClass() ) ) {
                return false;
            }
            ChunkID _ChunkID = ( ChunkID ) object;
            return ( ( this._ChunkColIndex == _ChunkID._ChunkColIndex ) && ( this._ChunkRowIndex == _ChunkID._ChunkRowIndex ) );
        }
        
        /**
         * Method required by Comparable
         */
        public int compareTo( Object object ) {
            //try {
            // Cast failure should throw NullPointerException or ClassCastException
            ChunkID _ChunkID = ( ChunkID ) object;
            if ( _ChunkID._ChunkRowIndex > this._ChunkRowIndex ) {
                return 1;
            }
            if ( _ChunkID._ChunkRowIndex < this._ChunkRowIndex ) {
                return -1;
            }
            if ( _ChunkID._ChunkColIndex > this._ChunkColIndex ) {
                return 1;
            }
            if ( _ChunkID._ChunkColIndex < this._ChunkColIndex ) {
                return -1;
            }
            return 0;
            //} catch ( Exception e0 ) {
            //    return -1;
            //}
        }
        
    }
    
}